"""Constants which are related to Canadian composer Marc Sabat."""

from mutwo.parameters import pitches

TUNEABLE_INTERVAL_TO_DIFFICULTY = {
    pitches.JustIntonationPitch("1/1").exponents: 0,
    pitches.JustIntonationPitch("8/7").exponents: 1,
    pitches.JustIntonationPitch("7/6").exponents: 0,
    pitches.JustIntonationPitch("6/5").exponents: 0,
    pitches.JustIntonationPitch("11/9").exponents: 2,
    pitches.JustIntonationPitch("5/4").exponents: 0,
    pitches.JustIntonationPitch("9/7").exponents: 0,
    pitches.JustIntonationPitch("13/10").exponents: 1,
    pitches.JustIntonationPitch("4/3").exponents: 0,
    pitches.JustIntonationPitch("11/8").exponents: 2,
    pitches.JustIntonationPitch("7/5").exponents: 0,
    pitches.JustIntonationPitch("10/7").exponents: 1,
    pitches.JustIntonationPitch("13/9").exponents: 2,
    pitches.JustIntonationPitch("16/11").exponents: 2,
    pitches.JustIntonationPitch("3/2").exponents: 0,
    pitches.JustIntonationPitch("14/9").exponents: 2,
    pitches.JustIntonationPitch("11/7").exponents: 1,
    pitches.JustIntonationPitch("8/5").exponents: 0,
    pitches.JustIntonationPitch("13/8").exponents: 1,
    pitches.JustIntonationPitch("5/3").exponents: 0,
    pitches.JustIntonationPitch("12/7").exponents: 1,
    pitches.JustIntonationPitch("7/4").exponents: 0,
    pitches.JustIntonationPitch("9/5").exponents: 0,
    pitches.JustIntonationPitch("11/6").exponents: 1,
    pitches.JustIntonationPitch("13/7").exponents: 1,
    pitches.JustIntonationPitch("15/8").exponents: 2,
    pitches.JustIntonationPitch("23/12").exponents: 2,
    pitches.JustIntonationPitch("2/1").exponents: 0,
    pitches.JustIntonationPitch("13/6").exponents: 1,
    pitches.JustIntonationPitch("11/5").exponents: 1,
    pitches.JustIntonationPitch("9/4").exponents: 0,
    pitches.JustIntonationPitch("7/3").exponents: 0,
    pitches.JustIntonationPitch("19/8").exponents: 2,
    pitches.JustIntonationPitch("12/5").exponents: 0,
    pitches.JustIntonationPitch("17/7").exponents: 1,
    pitches.JustIntonationPitch("5/2").exponents: 0,
    pitches.JustIntonationPitch("18/7").exponents: 1,
    pitches.JustIntonationPitch("13/5").exponents: 1,
    pitches.JustIntonationPitch("8/3").exponents: 0,
    pitches.JustIntonationPitch("11/4").exponents: 0,
    pitches.JustIntonationPitch("14/5").exponents: 0,
    pitches.JustIntonationPitch("17/6").exponents: 1,
    pitches.JustIntonationPitch("20/7").exponents: 1,
    pitches.JustIntonationPitch("23/8").exponents: 1,
    pitches.JustIntonationPitch("3/1").exponents: 0,
    pitches.JustIntonationPitch("28/9").exponents: 2,
    pitches.JustIntonationPitch("25/8").exponents: 2,
    pitches.JustIntonationPitch("22/7").exponents: 2,
    pitches.JustIntonationPitch("19/6").exponents: 2,
    pitches.JustIntonationPitch("16/5").exponents: 0,
    pitches.JustIntonationPitch("13/4").exponents: 0,
    pitches.JustIntonationPitch("10/3").exponents: 0,
    pitches.JustIntonationPitch("27/8").exponents: 2,
    pitches.JustIntonationPitch("17/5").exponents: 1,
    pitches.JustIntonationPitch("24/7").exponents: 2,
    pitches.JustIntonationPitch("7/2").exponents: 0,
    pitches.JustIntonationPitch("18/5").exponents: 0,
    pitches.JustIntonationPitch("11/3").exponents: 0,
    pitches.JustIntonationPitch("15/4").exponents: 0,
    pitches.JustIntonationPitch("19/5").exponents: 1,
    pitches.JustIntonationPitch("23/6").exponents: 2,
    pitches.JustIntonationPitch("27/7").exponents: 2,
    pitches.JustIntonationPitch("4/1").exponents: 0,
    pitches.JustIntonationPitch("25/6").exponents: 2,
    pitches.JustIntonationPitch("21/5").exponents: 2,
    pitches.JustIntonationPitch("17/4").exponents: 1,
    pitches.JustIntonationPitch("13/3").exponents: 0,
    pitches.JustIntonationPitch("22/5").exponents: 2,
    pitches.JustIntonationPitch("9/2").exponents: 0,
    pitches.JustIntonationPitch("23/5").exponents: 1,
    pitches.JustIntonationPitch("14/3").exponents: 0,
    pitches.JustIntonationPitch("19/4").exponents: 1,
    pitches.JustIntonationPitch("24/5").exponents: 1,
    pitches.JustIntonationPitch("5/1").exponents: 0,
    pitches.JustIntonationPitch("26/5").exponents: 2,
    pitches.JustIntonationPitch("21/4").exponents: 1,
    pitches.JustIntonationPitch("16/3").exponents: 0,
    pitches.JustIntonationPitch("11/2").exponents: 0,
    pitches.JustIntonationPitch("28/5").exponents: 1,
    pitches.JustIntonationPitch("17/3").exponents: 0,
    pitches.JustIntonationPitch("23/4").exponents: 1,
    pitches.JustIntonationPitch("6/1").exponents: 0,
    pitches.JustIntonationPitch("25/4").exponents: 1,
    pitches.JustIntonationPitch("19/3").exponents: 0,
    pitches.JustIntonationPitch("13/2").exponents: 0,
    pitches.JustIntonationPitch("20/3").exponents: 0,
    pitches.JustIntonationPitch("27/4").exponents: 2,
    pitches.JustIntonationPitch("7/1").exponents: 0,
    pitches.JustIntonationPitch("22/3").exponents: 0,
    pitches.JustIntonationPitch("15/2").exponents: 0,
    pitches.JustIntonationPitch("23/3").exponents: 0,
    pitches.JustIntonationPitch("8/1").exponents: 0,
}
"""Tuneable Just Intonation Intervals sorted by difficulty, according to Marc Sabat."""

TUNEABLE_INTERVALS = tuple(
    pitches.JustIntonationPitch(exponents)
    for exponents in TUNEABLE_INTERVAL_TO_DIFFICULTY.keys()
)
"""Tuneable Just Intonation Intervals according to Marc Sabat."""



"""Module to quantize free :class:`SequentialEvent` to notation based abjad :class:`Container`"""

import abc
import typing
import warnings

try:
    import quicktions as fractions  # type: ignore
except ImportError:
    import fractions  # type: ignore

import abjad  # type: ignore
from abjadext import nauert  # type: ignore
from abjadext import rmakers  # type: ignore
import expenvelope  # type: ignore
import ranges  # type: ignore

from mutwo.converters import abc as converters_abc

from mutwo import events
from mutwo import parameters

from mutwo.utilities import tools

__all__ = (
    "SequentialEventToQuantizedAbjadContainerConverter",
    "ComplexSequentialEventToQuantizedAbjadContainerConverter",
    "ComplexSequentialEventToDurationLineBasedQuantizedAbjadContainerConverter",
    "FastSequentialEventToQuantizedAbjadContainerConverter",
    "FastSequentialEventToDurationLineBasedQuantizedAbjadContainerConverter",
)


class SequentialEventToQuantizedAbjadContainerConverter(converters_abc.Converter):
    """Quantize :class:`~mutwo.events.basic.SequentialEvent` objects.

    :param time_signatures: Set time signatures to divide the quantized abjad data
        in desired bar sizes. If the converted :class:`~mutwo.events.basic.SequentialEvent`
        is longer than the sum of all passed time signatures, the last time signature
        will be repeated for the remaining bars.
    :param tempo_envelope: Defines the tempo of the converted music. This is an
        :class:`expenvelope.Envelope` object which durations are beats and which
        levels are either numbers (that will be interpreted as beats per minute ('BPM'))
        or :class:`~mutwo.parameters.tempos.TempoPoint` objects. If no tempo envelope has
        been defined, Mutwo will assume a constant tempo of 1/4 = 120 BPM.
    """

    def __init__(
        self,
        time_signatures: typing.Sequence[abjad.TimeSignature] = (
            abjad.TimeSignature((4, 4)),
        ),
        tempo_envelope: expenvelope.Envelope = None,
    ):
        n_time_signatures = len(time_signatures)
        if n_time_signatures == 0:
            message = (
                "Found empty sequence for argument 'time_signatures'. Specify at least"
                " one time signature!"
            )
            raise ValueError(message)

        time_signatures = tuple(time_signatures)
        if tempo_envelope is None:
            tempo_envelope = expenvelope.Envelope.from_points(
                (0, parameters.tempos.TempoPoint(120)),
                (0, parameters.tempos.TempoPoint(120)),
            )

        self._time_signatures = time_signatures
        self._tempo_envelope = tempo_envelope

    @property
    def tempo_envelope(self) -> expenvelope.Envelope:
        return self._tempo_envelope

    # ###################################################################### #
    #               public methods for interaction with the user             #
    # ###################################################################### #

    @abc.abstractmethod
    def convert(
        self, sequential_event_to_convert: events.basic.SequentialEvent
    ) -> typing.Tuple[
        abjad.Container,
        typing.Tuple[typing.Tuple[typing.Tuple[int, ...], ...], ...],
    ]:
        raise NotImplementedError


class ComplexSequentialEventToQuantizedAbjadContainerConverter(
    SequentialEventToQuantizedAbjadContainerConverter
):
    """Quantize :class:`~mutwo.events.basic.SequentialEvent` objects via :mod:`abjadext.nauert`.

    :param time_signatures: Set time signatures to divide the quantized abjad data
        in desired bar sizes. If the converted :class:`~mutwo.events.basic.SequentialEvent`
        is longer than the sum of all passed time signatures, the last time signature
        will be repeated for the remaining bars.
    :param duration_unit: This defines the `duration_unit` of the passed
        :class:`~mutwo.events.basic.SequentialEvent` (how the
        :attr:`~mutwo.events.abc.Event.duration` attribute will be
        interpreted). Can either be 'beats' (default) or 'miliseconds'.
        WARNING: 'miliseconds' isn't working properly yet!
    :param tempo_envelope: Defines the tempo of the converted music. This is an
        :class:`expenvelope.Envelope` object which durations are beats and which
        levels are either numbers (that will be interpreted as beats per minute ('BPM'))
        or :class:`~mutwo.parameters.tempos.TempoPoint` objects. If no tempo envelope has
        been defined, Mutwo will assume a constant tempo of 1/4 = 120 BPM.
    :param attack_point_optimizer: Optionally the user can pass a
        :class:`nauert.AttackPointOptimizer` object. Attack point optimizer help to
        split events and tie them for better looking notation. The default attack point
        optimizer is :class:`nauert.MeasurewiseAttackPointOptimizer` which splits events
        to better represent metrical structures within bars. If no optimizer is desired
        this argument can be set to ``None``.

    Unlike :class:`FastSequentialEventToQuantizedAbjadContainerConverter` this converter
    supports nested tuplets and ties across tuplets. But this converter is much slower
    than the :class:`FastSequentialEventToQuantizedAbjadContainerConverter`. Because the
    converter depends on the abjad extension `nauert` its quality is dependent on the
    inner mechanism of the used package. Because the quantization made by the `nauert`
    package can be somewhat indeterministic a lot of tweaking may be necessary for
    complex musical structures.
    """

    # TODO(add proper miliseconds conversion: you will have to add the tempo_envelope
    # when building the QEventSequence. Furthermore you should auto write down the
    # metronome marks when initialising from miliseconds?)

    def __init__(
        self,
        time_signatures: typing.Sequence[abjad.TimeSignature] = (
            abjad.TimeSignature((4, 4)),
        ),
        duration_unit: str = "beats",  # for future: typing.Literal["beats", "miliseconds"]
        tempo_envelope: expenvelope.Envelope = None,
        attack_point_optimizer: typing.Optional[
            nauert.AttackPointOptimizer
        ] = nauert.MeasurewiseAttackPointOptimizer(),
        search_tree: typing.Optional[nauert.SearchTree] = None,
    ):
        if duration_unit == "miliseconds":
            # warning for not well implemented miliseconds conversion

            message = (
                "The current implementation can't apply tempo changes for duration unit"
                " 'miliseconds' yet! Furthermore to quantize via duration_unit"
                " 'miliseconds' isn't well tested yet and may return unexpected"
                " results."
            )
            warnings.warn(message)

        time_signatures = tuple(time_signatures)
        # nauert will raise an error if there is only one time signature
        if len(time_signatures) == 1:
            time_signatures += time_signatures

        super().__init__(time_signatures, tempo_envelope)

        self._duration_unit = duration_unit
        self._attack_point_optimizer = attack_point_optimizer
        self._q_schema = (
            ComplexSequentialEventToQuantizedAbjadContainerConverter._make_q_schema(
                self._time_signatures, search_tree
            )
        )

    # ###################################################################### #
    #                          static methods                                #
    # ###################################################################### #

    @staticmethod
    def _get_respective_q_event_from_abjad_leaf(
        abjad_leaf: typing.Union[abjad.Rest, abjad.Note]
    ) -> typing.Optional[nauert.QEvent]:
        # TODO(improve ugly, heuristic, unreliable code)
        try:
            return abjad.get.indicators(abjad_leaf)[0]["q_events"][0]
        except TypeError:
            return None
        except KeyError:
            return None
        except IndexError:
            return None

    @staticmethod
    def _process_abjad_leaf(
        indices: typing.List[int],
        abjad_leaf: abjad.Leaf,
        related_abjad_leaves_per_simple_event: typing.List[
            typing.List[typing.Tuple[int, ...]]
        ],
        q_event_sequence: nauert.QEventSequence,
        has_tie: bool,
        index_of_previous_q_event: int,
    ) -> typing.Tuple[bool, int]:
        q_event = ComplexSequentialEventToQuantizedAbjadContainerConverter._get_respective_q_event_from_abjad_leaf(
            abjad_leaf
        )

        if q_event and type(q_event) != nauert.TerminalQEvent:
            nth_q_event = q_event_sequence.sequence.index(q_event)
            related_abjad_leaves_per_simple_event[nth_q_event].append(tuple(indices))
            index_of_previous_q_event = nth_q_event
        elif has_tie:
            related_abjad_leaves_per_simple_event[index_of_previous_q_event].append(
                tuple(indices)
            )
        # skip leaves without any links
        # else:
        #     related_abjad_leaves_per_simple_event.append([tuple(indices)])

        has_tie = abjad.get.has_indicator(abjad_leaf, abjad.Tie)

        return has_tie, index_of_previous_q_event

    @staticmethod
    def _process_tuplet(
        indices: typing.List[int],
        tuplet: abjad.Tuplet,
        related_abjad_leaves_per_simple_event: typing.List[
            typing.List[typing.Tuple[int, ...]]
        ],
        q_event_sequence: nauert.QEventSequence,
        has_tie: bool,
        index_of_previous_q_event: int,
    ) -> typing.Tuple[bool, int]:
        for (
            nth_abjad_leaf_or_tuplet,
            abjad_leaf_or_tuplet,
        ) in enumerate(tuplet):
            (
                has_tie,
                index_of_previous_q_event,
            ) = ComplexSequentialEventToQuantizedAbjadContainerConverter._process_abjad_leaf_or_tuplet(
                indices + [nth_abjad_leaf_or_tuplet],
                abjad_leaf_or_tuplet,
                related_abjad_leaves_per_simple_event,
                q_event_sequence,
                has_tie,
                index_of_previous_q_event,
            )

        return has_tie, index_of_previous_q_event

    @staticmethod
    def _process_abjad_leaf_or_tuplet(
        indices: typing.List[int],
        abjad_leaf_or_tuplet: typing.Union[abjad.Tuplet, abjad.Leaf],
        related_abjad_leaves_per_simple_event: typing.List[
            typing.List[typing.Tuple[int, ...]]
        ],
        q_event_sequence: nauert.QEventSequence,
        has_tie: bool,
        index_of_previous_q_event: int,
    ) -> typing.Tuple[bool, int]:
        if isinstance(abjad_leaf_or_tuplet, abjad.Tuplet):
            return ComplexSequentialEventToQuantizedAbjadContainerConverter._process_tuplet(
                indices,
                abjad_leaf_or_tuplet,
                related_abjad_leaves_per_simple_event,
                q_event_sequence,
                has_tie,
                index_of_previous_q_event,
            )

        else:
            return ComplexSequentialEventToQuantizedAbjadContainerConverter._process_abjad_leaf(
                indices,
                abjad_leaf_or_tuplet,
                related_abjad_leaves_per_simple_event,
                q_event_sequence,
                has_tie,
                index_of_previous_q_event,
            )

    @staticmethod
    def _make_related_abjad_leaves_per_simple_event(
        sequential_event: events.basic.SequentialEvent,
        q_event_sequence: nauert.QEventSequence,
        quanitisized_abjad_leaves: abjad.Voice,
    ) -> typing.Tuple[typing.Tuple[typing.Tuple[int, ...], ...], ...,]:
        has_tie = False
        index_of_previous_q_event: int = 0
        related_abjad_leaves_per_simple_event: typing.List[
            typing.List[typing.Tuple[int, ...]]
        ] = [[] for _ in sequential_event]
        for nth_bar, bar in enumerate(quanitisized_abjad_leaves):
            for nth_abjad_leaf_or_tuplet, abjad_leaf_or_tuplet in enumerate(bar):
                (
                    has_tie,
                    index_of_previous_q_event,
                ) = ComplexSequentialEventToQuantizedAbjadContainerConverter._process_abjad_leaf_or_tuplet(
                    [nth_bar, nth_abjad_leaf_or_tuplet],
                    abjad_leaf_or_tuplet,
                    related_abjad_leaves_per_simple_event,
                    q_event_sequence,
                    has_tie,
                    index_of_previous_q_event,
                )

        return tuple(
            tuple(tuple(item) for item in pair)
            for pair in related_abjad_leaves_per_simple_event
        )

    @staticmethod
    def _make_q_schema(
        time_signatures: typing.Tuple[abjad.TimeSignature, ...],
        search_tree: typing.Optional[nauert.SearchTree],
    ) -> nauert.QSchema:
        formated_time_signatures = []
        for time_signature in time_signatures:
            formated_time_signatures.append({"time_signature": time_signature})

        keyword_arguments = {
            "use_full_measure": True,
            "tempo": abjad.MetronomeMark((1, 4), 60),
        }

        if search_tree:
            keyword_arguments.update({"search_tree": search_tree})

        return nauert.MeasurewiseQSchema(*formated_time_signatures, **keyword_arguments)

    # ###################################################################### #
    #                         private methods                                #
    # ###################################################################### #

    def _sequential_event_to_q_event_sequence(
        self, sequential_event: events.basic.SequentialEvent
    ) -> nauert.QEventSequence:
        durations = list(sequential_event.get_parameter("duration"))

        for nth_simple_event, simple_event in enumerate(sequential_event):
            if simple_event.is_rest:
                durations[nth_simple_event] = -durations[nth_simple_event]

        if self._duration_unit == "beats":
            return nauert.QEventSequence.from_tempo_scaled_durations(
                durations, tempo=abjad.MetronomeMark((1, 4), 60)
            )

        elif self._duration_unit == "miliseconds":
            return nauert.QEventSequence.from_millisecond_durations(durations)

        else:
            message = (
                "Unknown duration unit '{}'. Use duration unit 'beats' or"
                " 'miliseconds'.".format(self._duration_unit)
            )
            raise NotImplementedError(message)

    def _q_event_sequence_to_quanitisized_abjad_leaves(
        self, q_event_sequence: nauert.QEventSequence
    ) -> abjad.Voice:
        quantizer = nauert.Quantizer()
        return quantizer(
            q_event_sequence,
            q_schema=self._q_schema,
            attach_tempos=True if self._duration_unit == "miliseconds" else False,
            attack_point_optimizer=self._attack_point_optimizer,
        )

    # ###################################################################### #
    #               public methods for interaction with the user             #
    # ###################################################################### #

    def convert(
        self, sequential_event_to_convert: events.basic.SequentialEvent
    ) -> typing.Tuple[
        abjad.Container,
        typing.Tuple[typing.Tuple[typing.Tuple[int, ...], ...], ...],
    ]:
        q_event_sequence = self._sequential_event_to_q_event_sequence(
            sequential_event_to_convert
        )
        quanitisized_abjad_leaves = self._q_event_sequence_to_quanitisized_abjad_leaves(
            q_event_sequence
        )

        related_abjad_leaves_per_simple_event = ComplexSequentialEventToQuantizedAbjadContainerConverter._make_related_abjad_leaves_per_simple_event(
            sequential_event_to_convert, q_event_sequence, quanitisized_abjad_leaves
        )
        return (
            quanitisized_abjad_leaves,
            related_abjad_leaves_per_simple_event,
        )


class FastSequentialEventToQuantizedAbjadContainerConverter(
    SequentialEventToQuantizedAbjadContainerConverter
):
    """Quantize :class:`~mutwo.events.basic.SequentialEvent` object via abjad rmakers

    :param time_signatures: Set time signatures to divide the quantized abjad data
        in desired bar sizes. If the converted
        :class:`~mutwo.events.basic.SequentialEvent` is longer than the sum of
        all passed time signatures, the last time signature
        will be repeated for the remaining bars.
    :param tempo_envelope: Defines the tempo of the converted music. This is an
        :class:`expenvelope.Envelope` object which durations are beats and which
        levels are either numbers (that will be interpreted as beats per minute ('BPM'))
        or :class:`~mutwo.parameters.tempos.TempoPoint` objects. If no tempo envelope
        has been defined, Mutwo will assume a constant tempo of 1/4 = 120 BPM.

    This method is significantly faster than the
    :class:`ComplexSequentialEventToQuantizedAbjadContainerConverter`. But it also
    has several known limitations:

        1. :class:`FastSequentialEventToQuantizedAbjadContainerConverter` doesn't
           support nested tuplets.
        2. :class:`FastSequentialEventToQuantizedAbjadContainerConverter` doesn't
           support ties across tuplets with different prolation (or across tuplets
           and not-tuplet notation). If ties are desired the user has to build them
           manually before passing the :class:`~mutwo.events.basic.SequentialEvent`
           to the converter.
    """

    _maximum_dot_count = 1

    def __init__(
        self, *args, do_rewrite_meter: bool = True, add_beams: bool = True, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._do_rewrite_meter = do_rewrite_meter
        self._add_beams = add_beams

    # ###################################################################### #
    #                       static private methods                           #
    # ###################################################################### #

    @staticmethod
    def _find_offset_inventory(meter: abjad.Meter) -> tuple[abjad.Offset, ...]:
        for nth_offset_inventory, offset_inventory in enumerate(
            depthwise_offset_inventory := meter.depthwise_offset_inventory
        ):
            difference = offset_inventory[1] - offset_inventory[0]
            if difference == fractions.Fraction(1, 4):
                return offset_inventory
            elif difference <= fractions.Fraction(1, 4):
                return depthwise_offset_inventory[nth_offset_inventory - 1]
        return offset_inventory

    @staticmethod
    def _add_explicit_beams(
        bar: abjad.Container, meter: abjad.Meter, global_offset: abjad.Offset
    ) -> None:
        offset_inventory = FastSequentialEventToQuantizedAbjadContainerConverter._find_offset_inventory(
            meter
        )
        leaf_offsets = []
        # don't attach beams on tuplets
        relevant_bar_items = filter(
            lambda leaf_or_tuplet: isinstance(leaf_or_tuplet, abjad.Leaf)
            and leaf_or_tuplet.written_duration < fractions.Fraction(1, 4),
            bar,
        )
        leaves = abjad.select(relevant_bar_items).leaves()
        for leaf in leaves:
            offset = abjad.get.timespan(leaf).start_offset - global_offset
            leaf_offsets.append(offset)

        beam_ranges = []
        for start, end in zip(offset_inventory, offset_inventory[1:]):
            area = ranges.Range(start, end)
            offsets = tuple(filter(lambda offset: offset in area, leaf_offsets))
            n_elements = len(offsets)
            is_start_in_leaves = start in offsets

            # make new beam range
            if is_start_in_leaves and n_elements > 1:
                new_beam_range = [
                    leaf_offsets.index(offsets[0]),
                    leaf_offsets.index(offsets[-1]),
                ]
                beam_ranges.append(new_beam_range)

        for beam_range in beam_ranges:
            start, stop = beam_range
            abjad.attach(abjad.StartBeam(), leaves[start])
            abjad.attach(abjad.StopBeam(), leaves[stop])

        global_offset += offset_inventory[-1]
        return global_offset

    @staticmethod
    def _find_tuplet_indices(bar: abjad.Container) -> tuple[int, ...]:
        tuplet_indices = []
        for index, leaf_or_tuplet in enumerate(bar):
            if isinstance(leaf_or_tuplet, abjad.Tuplet):
                tuplet_indices.append(index)

        return tuple(tuplet_indices)

    @staticmethod
    def _group_tuplet_indices(tuplet_indices: tuple[int, ...]) -> list[list[int]]:
        """Put adjacent tuplet indices into groups."""

        grouped_tuplet_indices = [[]]
        last_tuplet_index = None
        for tuplet_index in tuplet_indices:
            if last_tuplet_index:
                difference = tuplet_index - last_tuplet_index
                if difference == 1:
                    grouped_tuplet_indices[-1].append(tuplet_index)
                else:
                    grouped_tuplet_indices.append([tuplet_index])
            else:
                grouped_tuplet_indices[-1].append(tuplet_index)
            last_tuplet_index = tuplet_index
        return grouped_tuplet_indices

    @staticmethod
    def _concatenate_adjacent_tuplets_for_one_group(
        bar: abjad.Container, group: list[int]
    ):
        implied_prolations = [bar[index].implied_prolation for index in group]
        common_prolation_groups = [[implied_prolations[0], [group[0]]]]
        for index, prolation in zip(group[1:], implied_prolations[1:]):
            if prolation == common_prolation_groups[-1][0]:
                common_prolation_groups[-1][1].append(index)
            else:
                common_prolation_groups.append([prolation, [index]])

        tuplets = []
        for prolation, tuplet_indices in common_prolation_groups:
            tuplet = abjad.Tuplet(prolation)
            for tuplet_index in tuplet_indices:
                for component in bar[tuplet_index]:
                    tuplet.append(abjad.mutate.copy(component))
            tuplets.append(tuplet)

        bar[group[0] : group[-1] + 1] = tuplets

    # ###################################################################### #
    #                       private methods                                  #
    # ###################################################################### #

    def _make_notes(
        self, sequential_event_to_convert: events.basic.SequentialEvent
    ) -> abjad.Selection:
        stack = rmakers.stack(
            rmakers.note(),
            rmakers.force_rest(
                lambda _: abjad.select(_)
                .logical_ties()
                .get(
                    [
                        index
                        for index, event in enumerate(sequential_event_to_convert)
                        if event.is_rest
                    ]
                ),
            ),
        )
        notes = stack(
            tuple(
                map(
                    lambda duration: (
                        fractions.Fraction(duration).numerator,
                        fractions.Fraction(duration).denominator,
                    ),
                    sequential_event_to_convert.get_parameter("duration"),
                )
            )
        )
        return notes

    def _concatenate_adjacent_tuplets_for_one_bar(self, bar: abjad.Container):
        tuplet_indices = (
            FastSequentialEventToQuantizedAbjadContainerConverter._find_tuplet_indices(
                bar
            )
        )
        if tuplet_indices:
            grouped_tuplet_indices = FastSequentialEventToQuantizedAbjadContainerConverter._group_tuplet_indices(
                tuplet_indices
            )
            for group in reversed(grouped_tuplet_indices):
                if len(group) > 1:
                    FastSequentialEventToQuantizedAbjadContainerConverter._concatenate_adjacent_tuplets_for_one_group(
                        bar, group
                    )

    def _concatenate_adjacent_tuplets(self, voice: abjad.Voice) -> abjad.Voice:
        for bar in voice:
            self._concatenate_adjacent_tuplets_for_one_bar(bar)

    def _rewrite_meter(self, voice: abjad.Voice):
        time_signatures = iter(self._time_signatures)
        last_time_signature = self._time_signatures[-1]
        # rewrite by meter
        global_offset = abjad.Offset((0, 1))
        previous_time_signature = None
        for bar in voice:
            try:
                time_signature = next(time_signatures)
            except StopIteration:
                time_signature = last_time_signature
            if time_signature != previous_time_signature:
                abjad.attach(time_signature, abjad.get.leaf(bar, 0))
            meter = abjad.Meter(time_signature)
            abjad.Meter.rewrite_meter(
                bar[:], time_signature, maximum_dot_count=self._maximum_dot_count
            )
            if self._add_beams:
                global_offset = self._add_explicit_beams(bar, meter, global_offset)
            previous_time_signature = time_signature

        last_bar = bar
        difference = time_signature.duration - abjad.get.duration(last_bar)
        if difference:

            stack = rmakers.stack(
                rmakers.note(),
                rmakers.force_rest(lambda _: abjad.select(_).logical_ties()[:]),
            )
            last_bar.extend(stack((difference,)))
            abjad.Meter.rewrite_meter(
                last_bar[:], time_signature, maximum_dot_count=self._maximum_dot_count
            )

    def _make_voice(
        self, sequential_event_to_convert: events.basic.SequentialEvent
    ) -> abjad.Voice:
        # first build notes
        notes = self._make_notes(sequential_event_to_convert)

        # split notes by time signatures
        notes_split_by_time_signatures = abjad.mutate.split(
            notes,
            [time_signature.duration for time_signature in self._time_signatures],
            cyclic=True,
        )
        bars = []
        for selection in notes_split_by_time_signatures:
            try:
                bar = abjad.Container(selection.items, simultaneous=False)
            except Exception:
                bar = abjad.Container(
                    abjad.mutate.copy(selection).items, simultaneous=False
                )
            bars.append(bar)
        voice = abjad.Voice(bars)
        if self._do_rewrite_meter:
            self._rewrite_meter(voice)
        self._concatenate_adjacent_tuplets(voice)
        return voice

    def _get_data_for_leaf(
        self, indices: typing.Tuple[int, ...], leaf: abjad.Leaf
    ) -> typing.Tuple[typing.Tuple[int, ...], bool, bool]:
        has_tie = abjad.get.indicator(leaf, abjad.Tie())
        is_rest = (
            isinstance(leaf, abjad.Rest)
            or isinstance(leaf, abjad.MultimeasureRest)
            or isinstance(leaf, abjad.Skip)
        )
        return indices, has_tie, is_rest

    def _get_data_for_tuplet_or_leaf(
        self,
        indices: typing.Tuple[int, ...],
        leaf_or_tuplet: typing.Union[abjad.Leaf, abjad.Tuplet],
    ) -> typing.Tuple[typing.Tuple[typing.Tuple[int, ...], bool], ...]:
        if isinstance(leaf_or_tuplet, abjad.Leaf):
            return (self._get_data_for_leaf(indices, leaf_or_tuplet),)
        else:
            data_per_leaf_or_tuplet = []
            for nth_leaf_or_tuplet_of_tuplet, sub_leaf_or_tuplet in enumerate(
                leaf_or_tuplet
            ):
                data_per_leaf_or_tuplet.extend(
                    self._get_data_for_tuplet_or_leaf(
                        indices + (nth_leaf_or_tuplet_of_tuplet,), sub_leaf_or_tuplet
                    )
                )
            return tuple(data_per_leaf_or_tuplet)

    def _make_related_abjad_leaves_per_simple_event(
        self, voice: abjad.Voice
    ) -> typing.Tuple[typing.Tuple[typing.Tuple[int, ...], ...], ...]:
        data_per_tuplet_or_leaf = []
        for nth_bar, bar in enumerate(voice):
            for nth_leaf_or_tuplet, leaf_or_tuplet in enumerate(bar):
                data_per_tuplet_or_leaf.extend(
                    self._get_data_for_tuplet_or_leaf(
                        (nth_bar, nth_leaf_or_tuplet), leaf_or_tuplet
                    )
                )

        related_abjad_leaves_per_simple_event = []
        related_abjad_leaves = []
        was_previous_note_rest = None
        has_previous_tie = None
        for indices, has_tie, is_rest in data_per_tuplet_or_leaf:
            if has_previous_tie or all((was_previous_note_rest, is_rest)):
                related_abjad_leaves.append(indices)
            else:
                if related_abjad_leaves:
                    related_abjad_leaves_per_simple_event.append(
                        tuple(related_abjad_leaves)
                    )
                related_abjad_leaves = [indices]

            has_previous_tie = has_tie
            was_previous_note_rest = is_rest

        if related_abjad_leaves:
            related_abjad_leaves_per_simple_event.append(tuple(related_abjad_leaves))

        return tuple(related_abjad_leaves_per_simple_event)

    def convert(
        self, sequential_event_to_convert: events.basic.SequentialEvent
    ) -> typing.Tuple[
        abjad.Container,
        typing.Tuple[typing.Tuple[typing.Tuple[int, ...], ...], ...],
    ]:
        voice = self._make_voice(sequential_event_to_convert)
        related_abjad_leaves_per_simple_event = (
            self._make_related_abjad_leaves_per_simple_event(voice)
        )
        return voice, related_abjad_leaves_per_simple_event


class ComplexSequentialEventToDurationLineBasedQuantizedAbjadContainerConverter(
    ComplexSequentialEventToQuantizedAbjadContainerConverter
):
    """Quantize :class:`~mutwo.events.basic.SequentialEvent` objects via :mod:`abjadext.nauert`.

    :param time_signatures: Set time signatures to divide the quantized abjad data
        in desired bar sizes. If the converted :class:`~mutwo.events.basic.SequentialEvent`
        is longer than the sum of all passed time signatures, the last time signature
        will be repeated for the remaining bars.
    :param duration_unit: This defines the `duration_unit` of the passed
        :class:`~mutwo.events.basic.SequentialEvent` (how the
        :attr:`~mutwo.events.abc.Event.duration` attribute will be
        interpreted). Can either be 'beats' (default) or 'miliseconds'.
    :param tempo_envelope: Defines the tempo of the converted music. This is an
        :class:`expenvelope.Envelope` object which durations are beats and which
        levels are either numbers (that will be interpreted as beats per minute ('BPM'))
        or :class:`~mutwo.parameters.tempos.TempoPoint` objects. If no tempo envelope has
        been defined, Mutwo will assume a constant tempo of 1/4 = 120 BPM.
    :param attack_point_optimizer: Optionally the user can pass a
        :class:`nauert.AttackPointOptimizer` object. Attack point optimizer help to
        split events and tie them for better looking notation. The default attack point
        optimizer is :class:`nauert.MeasurewiseAttackPointOptimizer` which splits events
        to better represent metrical structures within bars. If no optimizer is desired
        this argument can be set to ``None``.
    :param duration_line_minimum_length: The minimum length of a duration line.
    :type duration_line_minimum_length: int
    :param duration_line_thickness: The thickness of a duration line.
    :type duration_line_thickness: int

    This converter differs from
    :class:`SequentialEventToQuantizedAbjadContainerConverter` through
    the usage of duration lines for indicating rhythm instead of using
    flags, beams, dots and note head colors.

    **Note**:

    Don't forget to add the 'Duration_line_engraver' to the resulting
    abjad Voice, otherwise Lilypond won't be able to render the desired output.

    **Example:**

    >>> import abjad
    >>> from mutwo.converters.frontends import abjad as mutwo_abjad
    >>> from mutwo.events import basic, music
    >>> converter = frontends.abjad.SequentialEventToAbjadVoiceConverter(
    >>>     frontends.abjad.ComplexSequentialEventToDurationLineBasedQuantizedAbjadContainerConverter(
    >>>        )
    >>>    )
    >>> sequential_event_to_convert = basic.SequentialEvent(
    >>>     [music.NoteLike("c", 0.125), music.NoteLike("d", 1), music.NoteLike([], 0.125), music.NoteLike("e", 0.16666), music.NoteLike("e", 0.08333333333333333)]
    >>>    )
    >>> converted_sequential_event = converter.convert(sequential_event_to_convert)
    >>> converted_sequential_event.consists_commands.append("Duration_line_engraver")
    """

    def __init__(
        self,
        time_signatures: typing.Sequence[abjad.TimeSignature] = (
            abjad.TimeSignature((4, 4)),
        ),
        duration_unit: str = "beats",  # for future: typing.Literal["beats", "miliseconds"]
        tempo_envelope: expenvelope.Envelope = None,
        attack_point_optimizer: typing.Optional[
            nauert.AttackPointOptimizer
        ] = nauert.MeasurewiseAttackPointOptimizer(),
        search_tree: typing.Optional[nauert.SearchTree] = None,
        duration_line_minimum_length: int = 6,
        duration_line_thickness: int = 3,
    ):
        super().__init__(
            time_signatures,
            duration_unit,
            tempo_envelope,
            attack_point_optimizer,
            search_tree,
        )
        self._duration_line_minimum_length = duration_line_minimum_length
        self._duration_line_thickness = duration_line_thickness

    def _prepare_first_element(self, first_element: abjad.Leaf):
        # don't write rests (simply write empty space)
        abjad.attach(abjad.LilyPondLiteral("\\omit Staff.Rest"), first_element)
        abjad.attach(
            abjad.LilyPondLiteral("\\omit Staff.MultiMeasureRest"), first_element
        )
        # don't write stems (Rhythm get defined by duration line)
        abjad.attach(abjad.LilyPondLiteral("\\omit Staff.Stem"), first_element)
        # don't write flags (Rhythm get defined by duration line)
        abjad.attach(abjad.LilyPondLiteral("\\omit Staff.Flag"), first_element)
        # don't write beams (Rhythm get defined by duration line)
        abjad.attach(abjad.LilyPondLiteral("\\omit Staff.Beam"), first_element)
        # don't write dots (Rhythm get defined by duration line)
        abjad.attach(
            abjad.LilyPondLiteral("\\override Staff.Dots.dot-count = #0"),
            first_element,
        )
        # only write black note heads (Rhythm get defined by duration line)
        abjad.attach(
            abjad.LilyPondLiteral("\\override Staff.NoteHead.duration-log = 2"),
            first_element,
        )
        # set duration line properties
        abjad.attach(
            abjad.LilyPondLiteral(
                "\\override Staff.DurationLine.minimum-length = {}".format(
                    self._duration_line_minimum_length
                )
            ),
            first_element,
        )
        abjad.attach(
            abjad.LilyPondLiteral(
                "\\override Staff.DurationLine.thickness = {}".format(
                    self._duration_line_thickness
                )
            ),
            first_element,
        )

    def _adjust_quantisized_abjad_leaves(
        self,
        quanitisized_abjad_leaves: abjad.Container,
        related_abjad_leaves_per_simple_event: typing.Tuple[
            typing.Tuple[typing.Tuple[int, ...], ...], ...
        ],
    ):
        is_first = True

        for abjad_leaves_indices in related_abjad_leaves_per_simple_event:
            if abjad_leaves_indices:
                first_element = tools.get_nested_item_from_indices(
                    abjad_leaves_indices[0], quanitisized_abjad_leaves
                )
                if is_first:
                    self._prepare_first_element(first_element)
                    is_first = False

                is_active = bool(abjad.get.pitches(first_element))
                if is_active:
                    if len(abjad_leaves_indices) > 1:
                        abjad.detach(abjad.Tie(), first_element)

                    abjad.attach(
                        abjad.LilyPondLiteral("\\-", format_slot="after"), first_element
                    )

                    for indices in abjad_leaves_indices[1:]:
                        element = tools.get_nested_item_from_indices(
                            indices, quanitisized_abjad_leaves
                        )
                        tools.set_nested_item_from_indices(
                            indices,
                            quanitisized_abjad_leaves,
                            abjad.Skip(element.written_duration),
                        )

    def convert(
        self, sequential_event_to_convert: events.basic.SequentialEvent
    ) -> typing.Tuple[
        abjad.Container,
        typing.Tuple[typing.Tuple[typing.Tuple[int, ...], ...], ...],
    ]:

        (
            quanitisized_abjad_leaves,
            related_abjad_leaves_per_simple_event,
        ) = super().convert(sequential_event_to_convert)

        self._adjust_quantisized_abjad_leaves(
            quanitisized_abjad_leaves, related_abjad_leaves_per_simple_event
        )

        return quanitisized_abjad_leaves, related_abjad_leaves_per_simple_event


class FastSequentialEventToDurationLineBasedQuantizedAbjadContainerConverter(
    FastSequentialEventToQuantizedAbjadContainerConverter
):
    """Quantize :class:`~mutwo.events.basic.SequentialEvent` objects via :mod:`abjadext.nauert`.

    :param time_signatures: Set time signatures to divide the quantized abjad data
        in desired bar sizes. If the converted :class:`~mutwo.events.basic.SequentialEvent`
        is longer than the sum of all passed time signatures, the last time signature
        will be repeated for the remaining bars.
    :param duration_unit: This defines the `duration_unit` of the passed
        :class:`~mutwo.events.basic.SequentialEvent` (how the
        :attr:`~mutwo.events.abc.Event.duration` attribute will be
        interpreted). Can either be 'beats' (default) or 'miliseconds'.
    :param tempo_envelope: Defines the tempo of the converted music. This is an
        :class:`expenvelope.Envelope` object which durations are beats and which
        levels are either numbers (that will be interpreted as beats per minute ('BPM'))
        or :class:`~mutwo.parameters.tempos.TempoPoint` objects. If no tempo envelope has
        been defined, Mutwo will assume a constant tempo of 1/4 = 120 BPM.
    :param attack_point_optimizer: Optionally the user can pass a
        :class:`nauert.AttackPointOptimizer` object. Attack point optimizer help to
        split events and tie them for better looking notation. The default attack point
        optimizer is :class:`nauert.MeasurewiseAttackPointOptimizer` which splits events
        to better represent metrical structures within bars. If no optimizer is desired
        this argument can be set to ``None``.
    :param duration_line_minimum_length: The minimum length of a duration line.
    :type duration_line_minimum_length: int
    :param duration_line_thickness: The thickness of a duration line.
    :type duration_line_thickness: int

    This converter differs from
    :class:`SequentialEventToQuantizedAbjadContainerConverter` through
    the usage of duration lines for indicating rhythm instead of using
    flags, beams, dots and note head colors.

    **Note**:

    Don't forget to add the 'Duration_line_engraver' to the resulting
    abjad Voice, otherwise Lilypond won't be able to render the desired output.

    **Example:**

    >>> import abjad
    >>> from mutwo.converters.frontends import abjad as mutwo_abjad
    >>> from mutwo.events import basic, music
    >>> converter = frontends.abjad.SequentialEventToAbjadVoiceConverter(
    >>>     frontends.abjad.ComplexSequentialEventToDurationLineBasedQuantizedAbjadContainerConverter(
    >>>        )
    >>>    )
    >>> sequential_event_to_convert = basic.SequentialEvent(
    >>>     [music.NoteLike("c", 0.125), music.NoteLike("d", 1), music.NoteLike([], 0.125), music.NoteLike("e", 0.16666), music.NoteLike("e", 0.08333333333333333)]
    >>>    )
    >>> converted_sequential_event = converter.convert(sequential_event_to_convert)
    >>> converted_sequential_event.consists_commands.append("Duration_line_engraver")
    """

    def __init__(
        self,
        *args,
        duration_line_minimum_length: int = 6,
        duration_line_thickness: int = 3,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._duration_line_minimum_length = duration_line_minimum_length
        self._duration_line_thickness = duration_line_thickness

    def _prepare_first_element(self, first_element: abjad.Leaf):
        # don't write rests (simply write empty space)
        abjad.attach(abjad.LilyPondLiteral("\\omit Staff.Rest"), first_element)
        abjad.attach(
            abjad.LilyPondLiteral("\\omit Staff.MultiMeasureRest"), first_element
        )
        # don't write stems (Rhythm get defined by duration line)
        abjad.attach(abjad.LilyPondLiteral("\\omit Staff.Stem"), first_element)
        # don't write flags (Rhythm get defined by duration line)
        abjad.attach(abjad.LilyPondLiteral("\\omit Staff.Flag"), first_element)
        # don't write beams (Rhythm get defined by duration line)
        abjad.attach(abjad.LilyPondLiteral("\\omit Staff.Beam"), first_element)
        # don't write dots (Rhythm get defined by duration line)
        abjad.attach(
            abjad.LilyPondLiteral("\\override Staff.Dots.dot-count = #0"),
            first_element,
        )
        # only write black note heads (Rhythm get defined by duration line)
        abjad.attach(
            abjad.LilyPondLiteral("\\override Staff.NoteHead.duration-log = 2"),
            first_element,
        )
        # set duration line properties
        abjad.attach(
            abjad.LilyPondLiteral(
                "\\override Staff.DurationLine.minimum-length = {}".format(
                    self._duration_line_minimum_length
                )
            ),
            first_element,
        )
        abjad.attach(
            abjad.LilyPondLiteral(
                "\\override Staff.DurationLine.thickness = {}".format(
                    self._duration_line_thickness
                )
            ),
            first_element,
        )

    def _adjust_quantisized_abjad_leaves(
        self,
        quanitisized_abjad_leaves: abjad.Container,
        related_abjad_leaves_per_simple_event: typing.Tuple[
            typing.Tuple[typing.Tuple[int, ...], ...], ...
        ],
    ):
        is_first = True

        for abjad_leaves_indices in related_abjad_leaves_per_simple_event:
            if abjad_leaves_indices:
                first_element = tools.get_nested_item_from_indices(
                    abjad_leaves_indices[0], quanitisized_abjad_leaves
                )
                if is_first:
                    self._prepare_first_element(first_element)
                    is_first = False

                is_active = bool(abjad.get.pitches(first_element))
                if is_active:
                    if len(abjad_leaves_indices) > 1:
                        abjad.detach(abjad.Tie(), first_element)

                    abjad.attach(
                        abjad.LilyPondLiteral("\\-", format_slot="after"), first_element
                    )

                    for indices in abjad_leaves_indices[1:]:
                        element = tools.get_nested_item_from_indices(
                            indices, quanitisized_abjad_leaves
                        )
                        tools.set_nested_item_from_indices(
                            indices,
                            quanitisized_abjad_leaves,
                            abjad.Skip(element.written_duration),
                        )

    def convert(
        self, sequential_event_to_convert: events.basic.SequentialEvent
    ) -> typing.Tuple[
        abjad.Container,
        typing.Tuple[typing.Tuple[typing.Tuple[int, ...], ...], ...],
    ]:

        (
            quanitisized_abjad_leaves,
            related_abjad_leaves_per_simple_event,
        ) = super().convert(sequential_event_to_convert)

        self._adjust_quantisized_abjad_leaves(
            quanitisized_abjad_leaves, related_abjad_leaves_per_simple_event
        )

        # only assign first item to abjad leaves
        post_processed_releated_abjad_leaves_per_simple_event = []
        for related_abjad_leaves in related_abjad_leaves_per_simple_event:
            post_processed_releated_abjad_leaves_per_simple_event.append(
                (related_abjad_leaves[0],)
            )

        return (
            quanitisized_abjad_leaves,
            post_processed_releated_abjad_leaves_per_simple_event,
        )



"""Module for routines which pre- or postprocess abjad containers."""

import abc
import datetime
import typing

import abjad

from mutwo import events
from mutwo import parameters
from mutwo.utilities import constants


class ProcessAbjadContainerRoutine(abc.ABC):
    @abc.abstractmethod
    def __call__(
        self,
        complex_event_to_convert: events.abc.ComplexEvent,
        container_to_process: abjad.Container,
    ):
        raise NotImplementedError


class AddDurationLineEngraver(ProcessAbjadContainerRoutine):
    def __call__(
        self,
        complex_event_to_convert: events.abc.ComplexEvent,
        container_to_process: abjad.Container,
    ):
        container_to_process.consists_commands.append("Duration_line_engraver")


class AddInstrumentName(ProcessAbjadContainerRoutine):
    def __init__(
        self,
        complex_event_to_instrument_name: typing.Callable[
            [events.abc.ComplexEvent], str
        ] = lambda complex_event: complex_event.instrument_name,
        complex_event_to_short_instrument_name: typing.Callable[
            [events.abc.ComplexEvent], str
        ] = lambda complex_event: complex_event.short_instrument_name,
        instrument_name_font_size: str = "teeny",
        short_instrument_name_font_size: str = "teeny",
    ):
        self._complex_event_to_instrument_name = complex_event_to_instrument_name
        self._complex_event_to_short_instrument_name = (
            complex_event_to_short_instrument_name
        )
        self._instrument_name_font_size = instrument_name_font_size
        self._short_instrument_name_font_size = short_instrument_name_font_size

    def __call__(
        self,
        complex_event_to_convert: events.abc.ComplexEvent,
        container_to_process: abjad.Container,
    ):
        first_leaf = abjad.get.leaf(container_to_process[0], 0)

        try:
            instrument_name = self._complex_event_to_instrument_name(
                complex_event_to_convert
            )
        except AttributeError:
            instrument_name = None

        try:
            short_instrument_name = self._complex_event_to_short_instrument_name(
                complex_event_to_convert
            )
        except AttributeError:
            short_instrument_name = None

        lilypond_context = container_to_process.lilypond_context.name

        if instrument_name:
            set_instrument_name_command = (
                f"\\set {lilypond_context}.instrumentName = \\markup {{ "
                f" \\{self._instrument_name_font_size} {{ {instrument_name} }} }}"
            )
            abjad.attach(
                abjad.LilyPondLiteral(set_instrument_name_command),
                first_leaf,
            )
        if short_instrument_name:
            set_short_instrument_name_command = (
                f"\\set {lilypond_context}.shortInstrumentName = \\markup {{ "
                f" \\{self._short_instrument_name_font_size} {{"
                f" {short_instrument_name} }} }}"
            )
            abjad.attach(
                abjad.LilyPondLiteral(set_short_instrument_name_command),
                first_leaf,
            )


class AddAccidentalStyle(ProcessAbjadContainerRoutine):
    def __init__(self, accidental_style: str):
        self._accidental_style = accidental_style

    def __call__(
        self,
        complex_event_to_convert: events.abc.ComplexEvent,
        container_to_process: abjad.Container,
    ):
        first_leaf = abjad.get.leaf(container_to_process[0], 0)

        if self._accidental_style:
            abjad.attach(
                abjad.LilyPondLiteral(f'\\accidentalStyle "{self._accidental_style}"'),
                first_leaf,
            )


class AddTimeBracketMarks(ProcessAbjadContainerRoutine):

    # ###################################################################### #
    #                     private static methods                             #
    # ###################################################################### #

    @staticmethod
    def _format_time(
        time: constants.Real,
        round_function: typing.Callable[[float], int] = round,
    ) -> str:
        return format(datetime.timedelta(seconds=round_function(time)))[2:]

    @staticmethod
    def _attach_time_bracket_mark(
        leaf_to_attach_to: abjad.Leaf,
        time_bracket_mark: abjad.RehearsalMark,
        format_slot: str,
    ):
        abjad.attach(
            abjad.LilyPondLiteral(format(time_bracket_mark), format_slot=format_slot),
            leaf_to_attach_to,
        )

    @staticmethod
    def _add_time_bracket_mark_for_time(
        leaf_to_attach_to: abjad.Leaf,
        time: parameters.abc.DurationType,
        format_slot: str,
    ):
        if format_slot == "after":
            hint = "end at: "
        else:
            hint = "\\hspace #-10 start at: "
        formated_time = AddTimeBracketMarks._format_time(time)
        time_bracket_mark = (
            f"\\mark \\markup {{ \\teeny {{ {hint} }} \\override #'(font-family ."
            f" typewriter) \\small {{ {formated_time} }} }}"
        )
        AddTimeBracketMarks._attach_time_bracket_mark(
            leaf_to_attach_to, time_bracket_mark, format_slot
        )

    @staticmethod
    def _add_time_bracket_mark_for_time_range(
        leaf_to_attach_to: abjad.Leaf,
        time_range: typing.Tuple[
            parameters.abc.DurationType, parameters.abc.DurationType
        ],
        format_slot: str,
    ):
        if format_slot == "after":
            hint = "\\hspace #-10 end in range: "
            hint = "\\hspace #-7"
        else:
            hint = "\\hspace #-10 start in range: "
            hint = "\\hspace #-10"

        formated_time_range = tuple(
            AddTimeBracketMarks._format_time(time) for time in time_range
        )
        time_bracket_mark = (
            f"\\mark \\markup {{ \\small {{ {hint} }} \\override #'(font-family ."
            f" typewriter) \\small {{ {formated_time_range[0]}"
            " } \n\\raise #0.55 \n\\teeny { \\concat { \\arrow-head #X #LEFT ##t"
            " \\draw-line #'(1 . 0) \\arrow-head #X #RIGHT ##t } }\n\\override"
            " #'(font-family . typewriter) \\small {"
            f" {formated_time_range[1]} }} }}"
        )
        AddTimeBracketMarks._attach_time_bracket_mark(
            leaf_to_attach_to, time_bracket_mark, format_slot
        )

    @staticmethod
    def _add_time_bracket_mark_for_time_or_time_range(
        leaf_to_attach_to: abjad.Leaf,
        time_or_time_range: events.time_brackets.TimeOrTimeRange,
        format_slot: str,
    ):
        if hasattr(time_or_time_range, "__getitem__"):
            AddTimeBracketMarks._add_time_bracket_mark_for_time_range(
                leaf_to_attach_to, time_or_time_range, format_slot
            )
        else:
            AddTimeBracketMarks._add_time_bracket_mark_for_time(
                leaf_to_attach_to, time_or_time_range, format_slot
            )

    # ###################################################################### #
    #                         private methods                                #
    # ###################################################################### #

    def __call__(
        self,
        complex_event_to_convert: events.time_brackets.TimeBracket,
        container_to_process: abjad.Container,
    ):
        first_leaf = abjad.get.leaf(container_to_process[0], 0)
        last_leaf = abjad.get.leaf(container_to_process[0], -1)
        for leaf_to_attach_to, time_or_time_range, format_slot in (
            (
                first_leaf,
                complex_event_to_convert.start_or_start_range,
                "before",
            ),
            (
                last_leaf,
                complex_event_to_convert.end_or_end_range,
                "after",
            ),
        ):
            # don't add the ending time range for a tempo based time bracket
            # (only rely on the tempo mark in this case!)
            if format_slot != "after" or not (
                isinstance(
                    complex_event_to_convert,
                    events.time_brackets.TempoBasedTimeBracket,
                )
                and not complex_event_to_convert.force_spanning_of_end_or_end_range
            ):
                AddTimeBracketMarks._add_time_bracket_mark_for_time_or_time_range(
                    leaf_to_attach_to, time_or_time_range, format_slot
                )


class SetStaffSize(ProcessAbjadContainerRoutine):
    def __init__(self, difference_of_size: int):
        self._difference_of_size = difference_of_size

    def __call__(
        self,
        complex_event_to_convert: events.abc.ComplexEvent,
        container_to_process: abjad.Container,
    ):
        first_leaf = abjad.get.leaf(container_to_process[0], 0)
        abjad.attach(
            abjad.LilyPondLiteral(
                f"\\magnifyStaff #(magstep {self._difference_of_size})",
                format_slot="before",
            ),
            first_leaf,
        )



"""Module for Cage-like time brackets.

Time brackets are events which are defined by unspecified start and end times.
"""

import typing

import numpy as np  # type: ignore

from mutwo.events import abc as events_abc
from mutwo.events import basic
from mutwo.events import brackets
from mutwo.parameters import tempos
from mutwo.utilities import constants
from mutwo.utilities import exceptions

__all__ = (
    "TimeBracket",
    "TempoBasedTimeBracket",
    "TimeBracketContainer",
)


TimeRange = typing.Tuple[constants.Real, constants.Real]
TimeOrTimeRange = typing.Union[constants.Real, TimeRange]

T = typing.TypeVar("T", bound=events_abc.Event)


class TimeBracket(brackets.Bracket, typing.Generic[T]):
    _class_specific_side_attributes = (
        "start_or_start_range",
        "end_or_end_range",
        "seed",
        "force_spanning_of_end_or_end_range",
    )

    def __init__(
        self,
        tagged_event_or_tagged_events: typing.Sequence[
            typing.Union[
                basic.TaggedSimpleEvent,
                basic.TaggedSequentialEvent,
                basic.TaggedSimultaneousEvent,
            ]
        ],
        start_or_start_range: TimeOrTimeRange,
        end_or_end_range: TimeOrTimeRange,
        seed: typing.Optional[int] = None,
        force_spanning_of_end_or_end_range: bool = True,
    ):
        super().__init__(
            tagged_event_or_tagged_events, start_or_start_range, end_or_end_range
        )

        time_bracket_random = np.random.default_rng(seed)
        self.force_spanning_of_end_or_end_range = force_spanning_of_end_or_end_range

        self._seed = seed
        self._time_bracket_random = time_bracket_random

    @staticmethod
    def _get_extrema_of_value_or_value_range(
        value_or_value_range: TimeOrTimeRange,
        operation: typing.Callable[[typing.Sequence], constants.Real],
    ):
        if isinstance(value_or_value_range, typing.Sequence):
            return operation(value_or_value_range)
        else:
            return value_or_value_range

    @staticmethod
    def _get_flexible_range_of_value_or_value_range(
        value_or_value_range: TimeOrTimeRange,
    ) -> float:
        if isinstance(value_or_value_range, typing.Sequence):
            return value_or_value_range[1] - value_or_value_range[0]
        else:
            return 0

    @staticmethod
    def _delay_value_or_value_range(
        value_or_value_range: TimeOrTimeRange, delay: constants.Real
    ) -> TimeOrTimeRange:
        if isinstance(value_or_value_range, typing.Sequence):
            return tuple(map(lambda value: value + delay, value_or_value_range))
        else:
            return value_or_value_range + delay

    @staticmethod
    def _extend_or_shrink_value_or_value_range(
        value_or_value_range: TimeOrTimeRange,
        difference: float,
        direction: typing.Literal["right", "left"],
    ) -> TimeOrTimeRange:
        if not isinstance(value_or_value_range, typing.Sequence):
            value_or_value_range = (value_or_value_range, value_or_value_range)

        new_value_or_value_range = list(value_or_value_range)
        if direction == "right":
            index = 1
        elif direction == "left":
            index = 0
        else:
            raise NotImplementedError(f"Unknown direction {direction}!")

        new_value_or_value_range[index] += difference

        if new_value_or_value_range[0] == new_value_or_value_range[1]:
            new_value_or_value_range = new_value_or_value_range[0]
        else:
            new_value_or_value_range = tuple(new_value_or_value_range)

        return new_value_or_value_range

    # ###################################################################### #
    #                           private methods                              #
    # ###################################################################### #

    def _assign_concrete_time(
        self, time_or_time_range: TimeOrTimeRange
    ) -> constants.Real:
        if hasattr(time_or_time_range, "__getitem__"):
            return self._time_bracket_random.uniform(*time_or_time_range)
        else:
            return time_or_time_range

    # ###################################################################### #
    #                         public properties                              #
    # ###################################################################### #

    @brackets.Bracket.start_or_start_range.setter
    def start_or_start_range(self, start_or_start_range: TimeOrTimeRange):
        self._start_or_start_range = start_or_start_range
        if hasattr(self, "_assigned_start_time"):
            self.assign_concrete_times()

    @brackets.Bracket.end_or_end_range.setter
    def end_or_end_range(self, end_or_end_range: TimeOrTimeRange):
        self._end_or_end_range = end_or_end_range
        if hasattr(self, "_assigned_end_time"):
            self.assign_concrete_times()

    @property
    def mean_start(self) -> constants.Real:
        return brackets.Bracket._get_mean_of_value_or_value_range(
            self.start_or_start_range
        )

    @property
    def mean_end(self) -> constants.Real:
        return TimeBracket._get_mean_of_value_or_value_range(self.end_or_end_range)

    @property
    def minimal_start(self) -> constants.Real:
        return TimeBracket._get_extrema_of_value_or_value_range(
            self.start_or_start_range, min
        )

    @property
    def maximum_end(self) -> constants.Real:
        return TimeBracket._get_extrema_of_value_or_value_range(
            self.end_or_end_range, max
        )

    @property
    def duration(self) -> constants.Real:
        return self.mean_end - self.mean_start

    @property
    def assigned_start_time(self) -> constants.Real:
        try:
            return self._assigned_start_time
        except AttributeError:
            raise exceptions.ValueNotAssignedError()

    @property
    def assigned_end_time(self) -> constants.Real:
        try:
            return self._assigned_end_time
        except AttributeError:
            raise exceptions.ValueNotAssignedError()

    @property
    def seed(self) -> typing.Optional[int]:
        return self._seed

    @property
    def as_active_event(self) -> basic.SequentialEvent[basic.SimpleEvent]:
        as_active_event = basic.SequentialEvent([])
        start, end = self.minimal_start, self.maximum_end
        if start > 0:
            inactive_event = basic.SimpleEvent(start)
            inactive_event.is_active = False
            as_active_event.append(inactive_event)
        active_event = basic.SimpleEvent(end - start)
        active_event.is_active = True
        as_active_event.append(active_event)
        return as_active_event

    @property
    def flexible_start_range(self) -> float:
        return TimeBracket._get_flexible_range_of_value_or_value_range(
            self.start_or_start_range
        )

    @flexible_start_range.setter
    def flexible_start_range(self, new_flexible_start_range: float):
        assert new_flexible_start_range >= 0
        difference = new_flexible_start_range - self.flexible_start_range
        self.start_or_start_range = self._extend_or_shrink_value_or_value_range(
            self.start_or_start_range, difference, "right"
        )

    @property
    def flexible_end_range(self) -> float:
        return TimeBracket._get_flexible_range_of_value_or_value_range(
            self.end_or_end_range
        )

    @flexible_end_range.setter
    def flexible_end_range(self, new_flexible_end_range: float):
        assert new_flexible_end_range >= 0
        difference = self.flexible_end_range - new_flexible_end_range
        self.end_or_end_range = self._extend_or_shrink_value_or_value_range(
            self.end_or_end_range, difference, "left"
        )

    # ###################################################################### #
    #                          public methods                                #
    # ###################################################################### #

    def assign_concrete_times(self):
        """Assign concrete values for start and end time"""

        self._assigned_start_time = self._assign_concrete_time(
            self.start_or_start_range
        )
        self._assigned_end_time = self._assign_concrete_time(self.end_or_end_range)

    def is_overlapping_with_other(self, other: "TimeBracket") -> bool:
        as_active_event = self.as_active_event
        as_active_event.cut_out(other.minimal_start, other.maximum_end)
        if as_active_event:
            for event in as_active_event:
                if event.is_active:
                    return True
        return False

    def delay(self, value: constants.Real):
        self._start_or_start_range = self._delay_value_or_value_range(
            self.start_or_start_range, value
        )
        self._end_or_end_range = self._delay_value_or_value_range(
            self.end_or_end_range, value
        )


class TempoBasedTimeBracket(TimeBracket, typing.Generic[T]):
    _class_specific_side_attributes = TimeBracket._class_specific_side_attributes + (
        "tempo",
    )

    def __init__(
        self,
        tagged_event_or_tagged_events: typing.Sequence[
            typing.Union[
                basic.TaggedSimpleEvent,
                basic.TaggedSequentialEvent,
                basic.TaggedSimultaneousEvent,
            ]
        ],
        start_or_start_range: TimeOrTimeRange,
        end_or_end_range: TimeOrTimeRange,
        tempo: typing.Union[constants.Real, tempos.TempoPoint],
        seed: typing.Optional[int] = None,
        force_spanning_of_end_or_end_range: bool = False,
    ):
        self.tempo = tempo
        super().__init__(
            tagged_event_or_tagged_events,
            start_or_start_range,
            end_or_end_range,
            seed,
            force_spanning_of_end_or_end_range,
        )


class TimeBracketContainer(brackets.BracketContainer[TimeBracket]):
    def delay(
        self,
        value: constants.Real,
        start: typing.Optional[constants.Real] = None,
        end: typing.Optional[constants.Real] = None,
        time_bracket_to_time_bracket_time: typing.Callable[
            [TimeBracket], constants.Real
        ] = lambda time_bracket: time_bracket.minimal_start,
    ):
        """Delay all time brackets in area from start to end by value.

        WARNING: This function is not safe (wrong usage could create overlapping
        time brackets with the same tags).
        """

        if not start:
            start = 0
        if not end:
            end = self[-1].maximum_end

        for time_bracket in self:
            time_bracket_time = time_bracket_to_time_bracket_time(time_bracket)
            if time_bracket_time >= start and time_bracket_time <= end:
                time_bracket.delay(value)

        self._resort()

    def register(
        self,
        bracket_to_register: T,
        test_for_overlapping_brackets: bool = True,
        tags_to_analyse: typing.Optional[typing.Tuple[str, ...]] = None,
    ):
        """Add new bracket to :class:`TimeBracketContainer`.

        :param bracket: The bracket which shall be added.
        :type bracket: TimeBracket
        """

        if test_for_overlapping_brackets:

            def get_tags_of_bracket(
                bracket_from_which_to_extract_tags: TimeBracket,
            ) -> typing.Tuple[str, ...]:
                return tuple(
                    tagged_event.tag
                    for tagged_event in bracket_from_which_to_extract_tags
                )

            # check for overlapping brackets before registering the new one
            if not tags_to_analyse:
                tags_to_analyse = get_tags_of_bracket(bracket_to_register)
            for registered_bracket in self._brackets:
                registered_bracket_tags = get_tags_of_bracket(registered_bracket)
                is_tag_in_other_tags = tuple(
                    tag in tags_to_analyse for tag in registered_bracket_tags
                )
                if any(is_tag_in_other_tags):
                    if registered_bracket.is_overlapping_with_other(
                        bracket_to_register
                    ):
                        raise exceptions.OverlappingTimeBracketsError()

        super().register(bracket_to_register)



import typing


from mutwo import converters
from mutwo import events
from mutwo import parameters

__all__ = ("ReaperMarkerConverter",)


class ReaperMarkerConverter(converters.abc.EventConverter):
    """Make Reaper Marker entries."""

    def __init__(
        self,
        simple_event_to_marker_name: typing.Callable[
            [events.basic.SimpleEvent], str
        ] = lambda simple_event: simple_event.name,  # type: ignore
        simple_event_to_marker_color: typing.Callable[
            [events.basic.SimpleEvent], str
        ] = lambda simple_event: simple_event.color,  # type: ignore
    ):
        self._simple_event_to_marker_name = simple_event_to_marker_name
        self._simple_event_to_marker_color = simple_event_to_marker_color

    def _convert_simple_event(
        self,
        simple_event: events.basic.SimpleEvent,
        absolute_entry_delay: parameters.abc.DurationType,
    ) -> typing.Tuple[str, ...]:
        try:
            marker_name = self._simple_event_to_marker_name(simple_event)
            marker_color = self._simple_event_to_marker_color(simple_event)
        except AttributeError:
            return tuple([])

        return ("{} {} {}".format(absolute_entry_delay, marker_name, marker_color),)

    def convert(self, event_to_convert: events.abc.Event) -> str:
        reaper_marker = tuple(
            "MARKER {} {}".format(nth_marker, marker_data)
            for nth_marker, marker_data in enumerate(
                self._convert_event(event_to_convert, 0)
            )
        )
        return "\n".join(reaper_marker)



"""Module for converting MMML text to Mutwo objects.

MMML is an abbreviation for 'Mutwos Music Markup Language'.
Similarly to `Music Macro Language <https://en.wikipedia.org/wiki/Music_Macro_Language>`_
it is intended to be a easy human readable and writeable plain text encoding
for musical data. The language is inspired by Lilypond, ABC Notation, Guido and Alda.
It differs from the former insofar as that the render engine (frontend) is
unspecified. Furthermore MMML is quite open to different notation specifications for
pitch, dynamics and rhythm as long as predefined identifier won't get overridden.
"""

import typing

from mutwo.converters import abc
from mutwo.converters.backends import mmml_constants
from mutwo import parameters

___all___ = (
    "MMMLSinglePitchConverter",
    "MMMLSingleJIPitchConverter",
    "MMMLPitchesConverter",
)


class MMMLSinglePitchConverter(abc.Converter):
    """Convert a single MMML pitch string to a mutwo pitch object."""

    def __init__(
        self,
        decodex_or_decodex_function: typing.Union[
            typing.Dict[str, parameters.abc.Pitch],
            typing.Callable[[str], parameters.abc.Pitch],
        ],
        octave_mark_processor: typing.Callable[
            [parameters.abc.Pitch, typing.Optional[str]], parameters.abc.Pitch
        ] = lambda pitch, _: pitch,
    ):
        self._decodex_or_decodex_function = decodex_or_decodex_function
        self._octave_mark_processor = octave_mark_processor

    def convert(
        self, mmml_pitch_to_convert: str
    ) -> typing.Optional[parameters.abc.Pitch]:
        mmml_pitch_class, *mmml_octave_mark = mmml_pitch_to_convert.split(
            mmml_constants.OCTAVE_IDENTIFIER
        )

        if mmml_pitch_class == mmml_constants.REST_IDENTIFIER:
            return None

        if mmml_octave_mark:
            mmml_octave_mark = mmml_octave_mark[0]

        if hasattr(self._decodex_or_decodex_function, "__call__"):
            mutwo_pitch = self._decodex_or_decodex_function(mmml_pitch_class)
        else:
            mutwo_pitch = self._decodex_or_decodex_function[mmml_pitch_class]

        mutwo_pitch = self._octave_mark_processor(mutwo_pitch, mmml_octave_mark)
        return mutwo_pitch


class MMMLSingleJIPitchConverter(MMMLSinglePitchConverter):
    def __init__(self):
        super().__init__(self._decodex_function, self._octave_mark_processor)

    @staticmethod
    def _get_and_add_prime_and_exponent_pair(
        prime_to_exponent: typing.Dict[int, int],
        element: str,
        current_base: str,
        current_exponent: int,
    ) -> typing.Tuple[str, int]:
        base = int(current_base)
        if base in prime_to_exponent:
            message = f"Found duplicate base {base}!"
            raise ValueError(message)
        prime_to_exponent.update({base: current_exponent})
        current_base = element
        current_exponent = 0
        is_separating = False
        return current_base, current_exponent, is_separating

    @staticmethod
    def _split_to_prime_and_exponent_pairs(
        mmml_pitch_class_to_convert: str,
    ) -> typing.Tuple[str, ...]:
        is_separating = False
        current_base = ""
        current_exponent = 0
        prime_to_exponent = {}
        for element in mmml_pitch_class_to_convert:
            if element.isdigit():
                if is_separating:
                    (
                        current_base,
                        current_exponent,
                        is_separating,
                    ) = MMMLSingleJIPitchConverter._get_and_add_prime_and_exponent_pair(
                        prime_to_exponent, element, current_base, current_exponent
                    )
                else:
                    current_base += element
            elif element == mmml_constants.JUST_INTONATION_POSITIVE_EXPONENT_IDENTIFIER:
                is_separating = True
                current_exponent += 1
            elif element == mmml_constants.JUST_INTONATION_NEGATIVE_EXPONENT_IDENTIFIER:
                is_separating = True
                current_exponent -= 1
            else:
                message = (
                    f"Found undefined sign {element} in {mmml_pitch_class_to_convert}!"
                )
                raise NotImplementedError(message)

        return prime_to_exponent

    @staticmethod
    def _decodex_function(
        mmml_pitch_class_to_convert: str,
    ) -> parameters.pitches.JustIntonationPitch:
        prime_to_exponent = MMMLSingleJIPitchConverter._split_to_prime_and_exponent_pairs(
            mmml_pitch_class_to_convert
        )
        numerator = 1
        denominator = 1

        for prime, exponent in prime_to_exponent.items():
            multiplied = prime ** abs(exponent)
            if exponent > 0:
                numerator *= multiplied
            else:
                denominator *= multiplied

        pitch = parameters.pitches.JustIntonationPitch(f"{numerator}/{denominator}")
        return pitch

    @staticmethod
    def _octave_mark_processor(
        just_intonation_pitch: parameters.pitches.JustIntonationPitch,
        mmml_octave_mark_to_apply: typing.Optional[str],
    ) -> parameters.pitches.JustIntonationPitch:
        if mmml_octave_mark_to_apply:
            octave = int(mmml_octave_mark_to_apply)
        else:
            octave = 0
        just_intonation_pitch.register(octave)
        return just_intonation_pitch


class MMMLPitchesConverter(abc.Converter):
    """Convert one or multiple MMML pitches to mutwo pitch objects."""

    def __init__(
        self,
        mmml_single_pitch_converter: MMMLSinglePitchConverter = MMMLSinglePitchConverter(
            lambda frequency: parameters.pitches.DirectPitch(float(frequency)),
            lambda pitch, octave: parameters.pitches.DirectPitch(
                pitch.frequency * (2 ** int(octave))
            )
            if octave
            else pitch,
        ),
        default_pitch: parameters.abc.Pitch = parameters.pitches.DirectPitch(440),
        default_octave_mark: str = "0",
    ):
        self._mmml_single_pitch_converter = mmml_single_pitch_converter
        self._default_pitch = default_pitch
        self._default_octave_mark = default_octave_mark

    @staticmethod
    def _get_octave_mark(mmml_pitch: str) -> typing.Optional[str]:
        _, *mmml_octave_mark = mmml_pitch.split(mmml_constants.OCTAVE_IDENTIFIER)
        if mmml_octave_mark:
            return mmml_octave_mark[0]
        return None

    def _convert_mmml_pitch_or_pitches_to_mutwo_pitch_or_pitches(
        self, mmml_pitch_or_pitches_to_convert: str, previous_octave_mark: str,
    ) -> typing.Tuple[typing.Tuple[parameters.abc.Pitch, ...], str]:
        converted_chord = []
        for mmml_pitch_to_convert in mmml_pitch_or_pitches_to_convert.split(
            mmml_constants.MULTIPLE_PITCHES_IDENTIFIER
        ):
            current_octave_mark = self._get_octave_mark(mmml_pitch_to_convert)
            if current_octave_mark:
                previous_octave_mark = current_octave_mark
            else:
                mmml_pitch_to_convert = (
                    f"{mmml_pitch_to_convert}:{previous_octave_mark}"
                )
            converted_pitch = self._mmml_single_pitch_converter.convert(
                mmml_pitch_to_convert
            )
            if converted_pitch:
                converted_chord.append(converted_pitch)
        converted_chord = tuple(converted_chord)
        return converted_chord, previous_octave_mark

    def convert(
        self,
        mmml_pitches_to_convert: typing.Union[
            str, typing.Sequence[typing.Optional[str]]
        ],
    ) -> typing.Tuple[typing.Tuple[parameters.abc.Pitch, ...], ...]:
        previous_chord = (self._default_pitch,)
        previous_octave_mark = self._default_octave_mark

        mmml_pitch_events_to_iterate = (
            mmml_pitches_to_convert.split(mmml_constants.EVENT_IDENTIFIER)
            if hasattr(mmml_pitches_to_convert, "split")
            else mmml_pitches_to_convert
        )

        converted_pitches = []
        for mmml_pitch_or_pitches_to_convert in mmml_pitch_events_to_iterate:
            if mmml_pitch_or_pitches_to_convert:
                (
                    converted_chord,
                    previous_octave_mark,
                ) = self._convert_mmml_pitch_or_pitches_to_mutwo_pitch_or_pitches(
                    mmml_pitch_or_pitches_to_convert, previous_octave_mark
                )
                previous_chord = converted_chord
            else:
                converted_chord = previous_chord
            converted_pitches.append(converted_chord)
        return tuple(converted_pitches)



from .quantization import *
from .building import *

from .quantization import __all__ as _all_quantization
from .building import __all__ as _all_building

__all__ = _all_quantization + _all_building



"""Convert :class:`mutwo.events.time_brackets.TimeBracket` to other data structures."""

import bisect
import itertools
import typing

import expenvelope  # type: ignore
import numpy as np  # type: ignore

from mutwo.converters import abc as converters_abc
from mutwo.converters.symmetrical import time_brackets_constants
from mutwo import events
from mutwo import parameters
from mutwo.utilities import constants
from mutwo.utilities import exceptions

__all__ = (
    "TimeBracketsToEventConverter",
    "EventToProbabilityCurveConverter",
)


class TimeBracketsToEventConverter(converters_abc.Converter):
    """Convert time bracket to events.basic.events.

    :param tag: Events with this tag will be extracted and converted.
    """

    def __init__(self, tag: str):
        self._tag = tag

    @staticmethod
    def _get_start_and_end_time(
        time_bracket: events.time_brackets.TimeBracket,
    ) -> typing.Tuple[constants.Real, constants.Real]:
        try:
            return (
                time_bracket.assigned_start_time,
                time_bracket.assigned_end_time,
            )
        except exceptions.ValueNotAssignedError:
            time_bracket.assign_concrete_times()
            return TimeBracketsToEventConverter._get_start_and_end_time(time_bracket)

    def _extract_tagged_event(
        self,
        events: typing.Sequence[
            typing.Union[
                events.basic.TaggedSimpleEvent,
                events.basic.TaggedSequentialEvent,
                events.basic.TaggedSimultaneousEvent,
            ]
        ],
    ) -> typing.Union[
        events.basic.TaggedSimpleEvent,
        events.basic.TaggedSequentialEvent,
        events.basic.TaggedSimultaneousEvent,
    ]:
        for tagged_event in events:
            if tagged_event.tag == self._tag:
                return tagged_event

        raise NotImplementedError(f"Couldn't find instrument '{self._tag}' in bracket.")

    def convert(
        self,
        time_brackets_to_convert: typing.Sequence[events.time_brackets.TimeBracket],
    ) -> typing.Tuple[
        typing.Union[
            events.basic.TaggedSimpleEvent,
            events.basic.TaggedSequentialEvent,
            events.basic.TaggedSimultaneousEvent,
        ],
        ...,
    ]:
        converted_events = []
        previous_end_time = 0
        for time_bracket in time_brackets_to_convert:
            (
                start_time,
                end_time,
            ) = TimeBracketsToEventConverter._get_start_and_end_time(time_bracket)

            # add rest if there is time between last time bracket and current bracket
            if previous_end_time < start_time:
                converted_events.append(
                    events.basic.TaggedSimpleEvent(
                        start_time - previous_end_time, tag=self._tag
                    )
                )
            # raise error if last bracket is still running while the current
            # time bracket already started
            elif previous_end_time > start_time:
                raise exceptions.OverlappingTimeBracketsError(
                    previous_end_time, start_time, time_bracket
                )

            duration_in_seconds = end_time - start_time
            tagged_event = self._extract_tagged_event(
                tuple(time_bracket)
            ).destructive_copy()
            tagged_event.duration = duration_in_seconds
            converted_events.append(tagged_event)

            previous_end_time = end_time

        return tuple(converted_events)


class EventToProbabilityCurveConverter(converters_abc.EventConverter):
    def __init__(
        self,
        start_or_start_range: events.time_brackets.TimeOrTimeRange,
        end_or_end_range: events.time_brackets.TimeOrTimeRange,
        time_grid: typing.Optional[float] = None,
        precision: typing.Optional[int] = None,
    ):
        if not time_grid:
            time_grid = time_brackets_constants.DEFAULT_TIME_GRID

        if not precision:
            precision = time_brackets_constants.DEFAULT_PRECISION

        self._time_grid = time_grid
        self._possible_start_times = (
            EventToProbabilityCurveConverter._get_possible_times_range(
                precision, start_or_start_range
            )
        )
        self._possible_end_times = (
            EventToProbabilityCurveConverter._get_possible_times_range(
                precision, end_or_end_range
            )
        )
        self._init_grid(start_or_start_range[0], end_or_end_range[1])

    def _init_grid(self, start, end):
        self._grid_size = int((end - start) // self._time_grid)
        self._grid = np.linspace(start, end, self._grid_size)

    @staticmethod
    def _get_possible_times_range(
        precision: int, time_or_time_range: events.time_brackets.TimeOrTimeRange
    ) -> typing.Tuple[float, ...]:
        try:
            possible_times = tuple(
                np.linspace(*time_or_time_range, num=precision, dtype=float)
            )
        except TypeError:
            possible_times = (time_or_time_range,)
        return possible_times

    def _get_probability_array_for_event(
        self,
        event_to_convert: events.basic.SimpleEvent,
        absolute_entry_delay: parameters.abc.DurationType,
    ) -> typing.Tuple[
        typing.Tuple[parameters.abc.DurationType, parameters.abc.DurationType]
    ]:
        start_and_end_times = np.array(
            [
                float(absolute_entry_delay),
                float(absolute_entry_delay + event_to_convert.duration),
            ]
        )

        probability_array = np.zeros(self._grid_size)
        for start_time, end_time in itertools.product(
            self._possible_start_times, self._possible_end_times
        ):
            if start_time < end_time:
                start_of_event, end_of_event = np.interp(
                    start_and_end_times,
                    (0, float(self._duration_of_event_to_convert)),
                    (float(start_time), float(end_time)),
                )
                start_index, end_index = tuple(
                    bisect.bisect_left(self._grid, position_of_event)
                    for position_of_event in (start_of_event, end_of_event)
                )
                probability_array[start_index:end_index] += 1

        # scale probability array
        probability_array = np.interp(
            probability_array, (0, np.max(probability_array)), (0, 1)
        )
        return probability_array

    def _convert_simple_event(
        self,
        event_to_convert: events.basic.SimpleEvent,
        absolute_entry_delay: parameters.abc.DurationType,
    ) -> typing.Sequence[typing.Any]:
        """Convert instance of :class:`mutwo.events.events.basic.SimpleEvent`.

        This method calculates the probability curve for a simple event. A higher number
        in the curve means that it is more likely that the converted event is active at
        the respective moment on the time line.
        """

        probability_array = self._get_probability_array_for_event(
            event_to_convert, absolute_entry_delay
        )

        probability_curve = expenvelope.Envelope.from_points(
            *tuple((time, value) for time, value in zip(self._grid, probability_array)),
        )
        return (probability_curve,)

    def convert(self, event_to_convert: events.abc.Event) -> typing.Any:
        self._duration_of_event_to_convert = event_to_convert.duration
        probability_curves = self._convert_event(event_to_convert, 0)
        return probability_curves



"""Build Lilypond scores via `Abjad <https://github.com/Abjad/abjad>`_ from Mutwo data.

The following converter classes help to quantize and translate Mutwo data to
Western notation. Due to the complex nature of this task, Mutwo tries to offer as
many optional arguments as possible through which the user can affect the conversion
routines. The most important class and best starting point for organising a conversion
setting is :class:`SequentialEventToAbjadVoiceConverter`.
If one wants to build complete scores from within mutwo, the module offers the
:class:`NestedComplexEventToAbjadContainerConverter`.
"""

from .parameters import *
from .events import *

from .parameters import __all__ as _all_parameters
from .events import __all__ as _all_events


__all__ = _all_parameters + _all_events



"""Module to build complex multi-level abjad based scores from mutwo events."""

import abc
import itertools
import typing

try:
    import quicktions as fractions  # type: ignore
except ImportError:
    import fractions  # type: ignore

import abjad  # type: ignore

from mutwo.converters import abc as converters_abc
from mutwo.converters.frontends import abjad_attachments
from mutwo.converters.frontends import abjad_constants
from mutwo.converters.frontends import abjad_process_container_routines

from mutwo import events
from mutwo import parameters

from mutwo.utilities import tools

from ..parameters import MutwoPitchToAbjadPitchConverter
from ..parameters import MutwoVolumeToAbjadAttachmentDynamicConverter
from ..parameters import TempoEnvelopeToAbjadAttachmentTempoConverter
from ..parameters import ComplexTempoEnvelopeToAbjadAttachmentTempoConverter

from .quantization import SequentialEventToQuantizedAbjadContainerConverter
from .quantization import ComplexSequentialEventToQuantizedAbjadContainerConverter
# from .quantization import FastSequentialEventToQuantizedAbjadContainerConverter
from .quantization import (
    ComplexSequentialEventToDurationLineBasedQuantizedAbjadContainerConverter,
)
from .quantization import (
    FastSequentialEventToDurationLineBasedQuantizedAbjadContainerConverter,
)


__all__ = (
    "ComplexEventToAbjadContainerConverter",
    "SequentialEventToAbjadVoiceConverter",
    "NestedComplexEventToAbjadContainerConverter",
    "NestedComplexEventToComplexEventToAbjadContainerConvertersConverter",
    "CycleBasedNestedComplexEventToComplexEventToAbjadContainerConvertersConverter",
    "TagBasedNestedComplexEventToComplexEventToAbjadContainerConvertersConverter",
)


class ComplexEventToAbjadContainerConverter(converters_abc.Converter):
    def __init__(
        self,
        abjad_container_class: typing.Type[abjad.Container],
        lilypond_type_of_abjad_container: str,
        complex_event_to_abjad_container_name: typing.Callable[
            [events.abc.ComplexEvent], str
        ],
        pre_process_abjad_container_routines: typing.Sequence[
            abjad_process_container_routines.ProcessAbjadContainerRoutine
        ],
        post_process_abjad_container_routines: typing.Sequence[
            abjad_process_container_routines.ProcessAbjadContainerRoutine
        ],
    ):
        self._abjad_container_class = abjad_container_class
        self._lilypond_type_of_abjad_container = lilypond_type_of_abjad_container
        self._complex_event_to_abjad_container_name = (
            complex_event_to_abjad_container_name
        )
        self._pre_process_abjad_container_routines = (
            pre_process_abjad_container_routines
        )
        self._post_process_abjad_container_routines = (
            post_process_abjad_container_routines
        )

    def _make_empty_abjad_container(
        self, complex_event_to_converter: events.abc.ComplexEvent
    ) -> abjad.Container:
        try:
            abjad_container_name = self._complex_event_to_abjad_container_name(
                complex_event_to_converter
            )
        except AttributeError:
            abjad_container_name = None
        return self._abjad_container_class(
            [],
            name=abjad_container_name,
            lilypond_type=self._lilypond_type_of_abjad_container,
            simultaneous=isinstance(
                complex_event_to_converter, events.basic.SimultaneousEvent
            ),
        )

    def _pre_process_abjad_container(
        self,
        complex_event_to_convert: events.abc.ComplexEvent,
        abjad_container_to_pre_process: abjad.Container,
    ):
        for (
            pre_process_abjad_container_routine
        ) in self._pre_process_abjad_container_routines:
            pre_process_abjad_container_routine(
                complex_event_to_convert, abjad_container_to_pre_process
            )

    def _post_process_abjad_container(
        self,
        complex_event_to_convert: events.abc.ComplexEvent,
        abjad_container_to_post_process: abjad.Container,
    ):
        for (
            post_process_abjad_container_routine
        ) in self._post_process_abjad_container_routines:
            post_process_abjad_container_routine(
                complex_event_to_convert, abjad_container_to_post_process
            )

    @abc.abstractmethod
    def _fill_abjad_container(
        self,
        abjad_container_to_fill: abjad.Container,
        complex_event_to_convert: events.abc.ComplexEvent,
    ):
        raise NotImplementedError

    def convert(
        self, complex_event_to_convert: events.abc.ComplexEvent
    ) -> abjad.Container:
        abjad_container = self._make_empty_abjad_container(complex_event_to_convert)
        self._pre_process_abjad_container(complex_event_to_convert, abjad_container)
        self._fill_abjad_container(abjad_container, complex_event_to_convert)
        self._post_process_abjad_container(complex_event_to_convert, abjad_container)
        return abjad_container


class SequentialEventToAbjadVoiceConverter(ComplexEventToAbjadContainerConverter):
    """Convert :class:`~mutwo.events.basic.SequentialEvent` to :class:`abjad.Voice`.

    :param sequential_event_to_quantized_abjad_container_converter: Class which
        defines how the Mutwo data will be quantized. See
        :class:`SequentialEventToQuantizedAbjadContainerConverter` for more information.
    :type sequential_event_to_quantized_abjad_container_converter: SequentialEventToQuantizedAbjadContainerConverter, optional
    :param simple_event_to_pitches: Function to extract from a
        :class:`mutwo.events.basic.SimpleEvent` a tuple that contains pitch objects
        (objects that inherit from :class:`mutwo.parameters.abc.Pitch`).
        By default it asks the Event for its
        :attr:`~mutwo.events.music.NoteLike.pitch_or_pitches` attribute
        (because by default :class:`mutwo.events.music.NoteLike` objects are expected).
        When using different Event classes than :class:`~mutwo.events.music.NoteLike`
        with a different name for their pitch property, this argument
        should be overridden.
        If the function call raises an :obj:`AttributeError` (e.g. if no pitch can be
        extracted), mutwo will assume an event without any pitches.
    :type simple_event_to_pitches: typing.Callable[[events.basic.SimpleEvent], parameters.abc.Pitch], optional
    :param simple_event_to_volume: Function to extract the volume from a
        :class:`mutwo.events.basic.SimpleEvent` in the purpose of generating dynamic
        indicators. The function should return an object that inherits from
        :class:`mutwo.parameters.abc.Volume`. By default it asks the Event for
        its :attr:`~mutwo.events.music.NoteLike.volume` attribute (because by default
        :class:`mutwo.events.music.NoteLike` objects are expected).
        When using different Event classes than :class:`~mutwo.events.music.NoteLike`
        with a different name for their volume property, this argument should
        be overridden.
        If the function call raises an :obj:`AttributeError` (e.g. if no volume can be
        extracted), mutwo will set :attr:`pitch_or_pitches` to an empty list and set
        volume to 0.
    :type simple_event_to_volume: typing.Callable[[events.basic.SimpleEvent], parameters.abc.Volume], optional
    :param simple_event_to_playing_indicators: Function to extract from a
        :class:`mutwo.events.basic.SimpleEvent` a
        :class:`mutwo.parameters.playing_indicators.PlayingIndicatorCollection`
        object. By default it asks the Event for its
        :attr:`~mutwo.events.music.NoteLike.playing_indicators`
        attribute (because by default :class:`mutwo.events.music.NoteLike`
        objects are expected).
        When using different Event classes than :class:`~mutwo.events.music.NoteLike`
        with a different name for their playing_indicators property, this argument
        should be overridden. If the
        function call raises an :obj:`AttributeError` (e.g. if no playing indicator
        collection can be extracted), mutwo will build a playing indicator collection
        from :const:`~mutwo.events.music_constants.DEFAULT_PLAYING_INDICATORS_COLLECTION_CLASS`.
    :type simple_event_to_playing_indicators: typing.Callable[[events.basic.SimpleEvent], parameters.playing_indicators.PlayingIndicatorCollection,], optional
    :param simple_event_to_notation_indicators: Function to extract from a
        :class:`mutwo.events.basic.SimpleEvent` a
        :class:`mutwo.parameters.notation_indicators.NotationIndicatorCollection`
        object. By default it asks the Event for its
        :attr:`~mutwo.events.music.NoteLike.notation_indicators`
        (because by default :class:`mutwo.events.music.NoteLike` objects are expected).
        When using different Event classes than ``NoteLike`` with a different name for
        their playing_indicators property, this argument should be overridden. If the
        function call raises an :obj:`AttributeError` (e.g. if no notation indicator
        collection can be extracted), mutwo will build a notation indicator collection
        from :const:`~mutwo.events.music_constants.DEFAULT_NOTATION_INDICATORS_COLLECTION_CLASS`
    :type simple_event_to_notation_indicators: typing.Callable[[events.basic.SimpleEvent], parameters.notation_indicators.NotationIndicatorCollection,], optional
    :param is_simple_event_rest: Function to detect if the
        the inspected :class:`mutwo.events.basic.SimpleEvent` is a Rest. By
        default Mutwo simply checks if 'pitch_or_pitches' contain any objects. If not,
        the Event will be interpreted as a rest.
    :type is_simple_event_rest: typing.Callable[[events.basic.SimpleEvent], bool], optional
    :param mutwo_pitch_to_abjad_pitch_converter: Class which defines how to convert
        :class:`mutwo.parameters.abc.Pitch` objects to :class:`abjad.Pitch` objects.
        See :class:`MutwoPitchToAbjadPitchConverter` for more information.
    :type mutwo_pitch_to_abjad_pitch_converter: MutwoPitchToAbjadPitchConverter, optional
    :param mutwo_volume_to_abjad_attachment_dynamic_converter: Class which defines how
        to convert :class:`mutwo.parameters.abc.Volume` objects to
        :class:`mutwo.converters.frontends.abjad_attachments.Dynamic` objects.
        See :class:`MutwoVolumeToAbjadAttachmentDynamicConverter` for more information.
    :type mutwo_volume_to_abjad_attachment_dynamic_converter: MutwoVolumeToAbjadAttachmentDynamicConverter, optional
    :param tempo_envelope_to_abjad_attachment_tempo_converter: Class which defines how
        to convert tempo envelopes to
        :class:`mutwo.converters.frontends.abjad_attachments.Tempo` objects.
        See :class:`TempoEnvelopeToAbjadAttachmentTempoConverter` for more information.
    :type tempo_envelope_to_abjad_attachment_tempo_converter: TempoEnvelopeToAbjadAttachmentTempoConverter, optional
    :param abjad_attachment_classes: A tuple which contains all available abjad attachment classes
        which shall be used by the converter.
    :type abjad_attachment_classes: typing.Sequence[abjad_attachments.AbjadAttachment], optional
    :param write_multimeasure_rests: Set to ``True`` if the converter should replace
        rests that last a complete bar with multimeasure rests (rests with uppercase
        "R" in Lilypond). Default to ``True``.
    :type write_multimeasure_rests: bool
    """

    def __init__(
        self,
        sequential_event_to_quantized_abjad_container_converter: SequentialEventToQuantizedAbjadContainerConverter = ComplexSequentialEventToQuantizedAbjadContainerConverter(),
        simple_event_to_pitches: typing.Callable[
            [events.basic.SimpleEvent], typing.List[parameters.abc.Pitch]
        ] = lambda simple_event: simple_event.pitch_or_pitches,  # type: ignore
        simple_event_to_volume: typing.Callable[
            [events.basic.SimpleEvent], parameters.abc.Volume
        ] = lambda simple_event: simple_event.volume,  # type: ignore
        simple_event_to_playing_indicators: typing.Callable[
            [events.basic.SimpleEvent],
            parameters.playing_indicators.PlayingIndicatorCollection,
        ] = lambda simple_event: simple_event.playing_indicators,  # type: ignore
        simple_event_to_notation_indicators: typing.Callable[
            [events.basic.SimpleEvent],
            parameters.notation_indicators.NotationIndicatorCollection,
        ] = lambda simple_event: simple_event.notation_indicators,  # type: ignore
        is_simple_event_rest: typing.Callable[[events.basic.SimpleEvent], bool] = None,
        mutwo_pitch_to_abjad_pitch_converter: MutwoPitchToAbjadPitchConverter = MutwoPitchToAbjadPitchConverter(),
        mutwo_volume_to_abjad_attachment_dynamic_converter: typing.Optional[
            MutwoVolumeToAbjadAttachmentDynamicConverter
        ] = MutwoVolumeToAbjadAttachmentDynamicConverter(),
        tempo_envelope_to_abjad_attachment_tempo_converter: typing.Optional[
            TempoEnvelopeToAbjadAttachmentTempoConverter
        ] = ComplexTempoEnvelopeToAbjadAttachmentTempoConverter(),
        abjad_attachment_classes: typing.Sequence[
            typing.Type[abjad_attachments.AbjadAttachment]
        ] = None,
        write_multimeasure_rests: bool = True,
        lilypond_type_of_abjad_container: str = "Voice",
        complex_event_to_abjad_container_name: typing.Callable[
            [events.abc.ComplexEvent], typing.Optional[str]
        ] = lambda _: None,
        pre_process_abjad_container_routines: typing.Sequence[
            abjad_process_container_routines.ProcessAbjadContainerRoutine
        ] = tuple([]),
        post_process_abjad_container_routines: typing.Sequence[
            abjad_process_container_routines.ProcessAbjadContainerRoutine
        ] = tuple([]),
    ):
        # special treatment for duration line based quantizer
        if isinstance(
            sequential_event_to_quantized_abjad_container_converter,
            (
                ComplexSequentialEventToDurationLineBasedQuantizedAbjadContainerConverter,
                FastSequentialEventToDurationLineBasedQuantizedAbjadContainerConverter,
            ),
        ):
            post_process_abjad_container_routines += (
                abjad_process_container_routines.AddDurationLineEngraver(),
            )

        super().__init__(
            abjad.Voice,
            lilypond_type_of_abjad_container,
            complex_event_to_abjad_container_name,
            pre_process_abjad_container_routines,
            post_process_abjad_container_routines,
        )

        if abjad_attachment_classes is None:
            abjad_attachment_classes = abjad_constants.DEFAULT_ABJAD_ATTACHMENT_CLASSES
        else:
            abjad_attachment_classes = tuple(abjad_attachment_classes)

        if is_simple_event_rest is None:

            def is_simple_event_rest(simple_event: events.basic.SimpleEvent) -> bool:
                try:
                    pitch_or_pitches = simple_event_to_pitches(simple_event)
                except AttributeError:
                    pitch_or_pitches = []

                return not bool(pitch_or_pitches)

        self._abjad_attachment_classes = abjad_attachment_classes

        self._available_abjad_attachments = tuple(
            abjad_attachment_class.get_class_name()
            for abjad_attachment_class in self._abjad_attachment_classes
        )

        self._sequential_event_to_quantized_abjad_container_converter = (
            sequential_event_to_quantized_abjad_container_converter
        )
        self._simple_event_to_pitches = simple_event_to_pitches
        self._simple_event_to_volume = simple_event_to_volume
        self._simple_event_to_playing_indicators = simple_event_to_playing_indicators
        self._simple_event_to_notation_indicators = simple_event_to_notation_indicators
        self._is_simple_event_rest = is_simple_event_rest
        self._mutwo_pitch_to_abjad_pitch_converter = (
            mutwo_pitch_to_abjad_pitch_converter
        )
        self._mutwo_volume_to_abjad_attachment_dynamic_converter = (
            mutwo_volume_to_abjad_attachment_dynamic_converter
        )
        if tempo_envelope_to_abjad_attachment_tempo_converter:
            tempo_attachments = tempo_envelope_to_abjad_attachment_tempo_converter.convert(
                self._sequential_event_to_quantized_abjad_container_converter.tempo_envelope
            )
        else:
            tempo_attachments = None
        self._tempo_attachments = tempo_attachments

        self._write_multimeasure_rests = write_multimeasure_rests

    # ###################################################################### #
    #                          static methods                                #
    # ###################################################################### #

    @staticmethod
    def _detect_abjad_event_type(pitches: typing.List[parameters.abc.Pitch]) -> type:
        n_pitches = len(pitches)
        if n_pitches == 0:
            abjad_event_type = abjad.Rest
        elif n_pitches == 1:
            abjad_event_type = abjad.Note
        else:
            abjad_event_type = abjad.Chord
        return abjad_event_type

    @staticmethod
    def _find_absolute_times_of_abjad_leaves(
        abjad_voice: abjad.Voice,
    ) -> typing.Tuple[fractions.Fraction, ...]:
        absolute_time_per_leaf: typing.List[fractions.Fraction] = []
        for leaf in abjad.select(abjad_voice).leaves():
            start, _ = abjad.get.timespan(leaf).offsets
            absolute_time_per_leaf.append(
                fractions.Fraction(start.numerator, start.denominator)
            )
        return tuple(absolute_time_per_leaf)

    @staticmethod
    def _replace_rests_with_full_measure_rests(abjad_voice: abjad.Voice) -> None:
        for bar in abjad_voice:
            if all((isinstance(item, abjad.Rest) for item in bar)):
                duration = sum((item.written_duration for item in bar))
                numerator, denominator = duration.numerator, duration.denominator
                abjad.mutate.replace(
                    bar[0],
                    abjad.MultimeasureRest(
                        abjad.Duration(1, denominator), multiplier=numerator
                    ),
                    wrappers=True,
                )
                del bar[1:]

    # ###################################################################### #
    #                          private methods                               #
    # ###################################################################### #

    def _indicator_collection_to_abjad_attachments(
        self,
        indicator_collection: parameters.abc.IndicatorCollection,
    ) -> typing.Dict[str, abjad_attachments.AbjadAttachment]:
        attachments = {}
        for abjad_attachment_class in self._abjad_attachment_classes:
            abjad_attachment = abjad_attachment_class.from_indicator_collection(
                indicator_collection
            )
            if abjad_attachment:
                attachments.update(
                    {abjad_attachment_class.get_class_name(): abjad_attachment}
                )

        return attachments

    def _volume_to_abjad_attachment(
        self, volume: parameters.abc.Volume
    ) -> typing.Dict[str, abjad_attachments.AbjadAttachment]:
        if self._mutwo_volume_to_abjad_attachment_dynamic_converter:
            abjad_attachment_dynamic = (
                self._mutwo_volume_to_abjad_attachment_dynamic_converter.convert(volume)
            )
            if abjad_attachment_dynamic:
                return {"dynamic": abjad_attachment_dynamic}
        return {}

    def _get_tempo_attachments_for_quantized_abjad_leaves(
        self,
        abjad_voice: abjad.Voice,
    ) -> typing.Tuple[
        typing.Tuple[
            int,
            typing.Union[
                abjad_attachments.Tempo, abjad_attachments.DynamicChangeIndicationStop
            ],
        ],
        ...,
    ]:
        absolute_time_per_leaf = (
            SequentialEventToAbjadVoiceConverter._find_absolute_times_of_abjad_leaves(
                abjad_voice
            )
        )

        assert absolute_time_per_leaf == tuple(sorted(absolute_time_per_leaf))

        leaf_index_to_tempo_attachment_pairs: typing.List[
            typing.Tuple[
                int,
                typing.Union[
                    abjad_attachments.Tempo,
                    abjad_attachments.DynamicChangeIndicationStop,
                ],
            ]
        ] = []
        for absolute_time, tempo_attachment in self._tempo_attachments:
            closest_leaf = tools.find_closest_index(
                absolute_time, absolute_time_per_leaf
            )
            # special case:
            # check for stop dynamic change indication
            # (has to applied to the previous leaf for
            #  better looking results)
            if tempo_attachment.stop_dynamic_change_indicaton:
                leaf_index_to_tempo_attachment_pairs.append(
                    (closest_leaf - 1, abjad_attachments.DynamicChangeIndicationStop())
                )
            leaf_index_to_tempo_attachment_pairs.append(
                (closest_leaf, tempo_attachment)
            )

        return tuple(leaf_index_to_tempo_attachment_pairs)

    def _get_attachments_for_quantized_abjad_leaves(
        self,
        extracted_data_per_simple_event: typing.Tuple[
            typing.Tuple[
                typing.List[parameters.abc.Pitch],
                parameters.abc.Volume,
                parameters.playing_indicators.PlayingIndicatorCollection,
                parameters.notation_indicators.NotationIndicatorCollection,
            ],
            ...,
        ],
    ) -> typing.Tuple[
        typing.Tuple[typing.Optional[abjad_attachments.AbjadAttachment], ...], ...
    ]:
        attachments_per_type_per_event: typing.Dict[
            str, typing.List[typing.Optional[abjad_attachments.AbjadAttachment]]
        ] = {
            attachment_name: [None for _ in extracted_data_per_simple_event]
            for attachment_name in self._available_abjad_attachments
        }
        for nth_event, extracted_data in enumerate(extracted_data_per_simple_event):
            _, volume, playing_indicators, notation_indicators = extracted_data
            attachments_for_nth_event = self._volume_to_abjad_attachment(volume)
            attachments_for_nth_event.update(
                self._indicator_collection_to_abjad_attachments(playing_indicators)
            )
            attachments_for_nth_event.update(
                self._indicator_collection_to_abjad_attachments(notation_indicators)
            )
            for attachment_name, attachment in attachments_for_nth_event.items():
                attachments_per_type_per_event[attachment_name][nth_event] = attachment

        return tuple(
            tuple(attachments)
            for attachments in attachments_per_type_per_event.values()
        )

    def _apply_tempos_on_quantized_abjad_leaves(
        self,
        quanitisized_abjad_leaves: abjad.Voice,
    ):
        if self._tempo_attachments:
            leaves = abjad.select(quanitisized_abjad_leaves).leaves()
            tempo_attachment_data = (
                self._get_tempo_attachments_for_quantized_abjad_leaves(
                    quanitisized_abjad_leaves
                )
            )
            for nth_event, tempo_attachment in tempo_attachment_data:
                tempo_attachment.process_leaves((leaves[nth_event],), None)

    def _apply_attachments_on_quantized_abjad_leaves(
        self,
        quanitisized_abjad_leaves: abjad.Voice,
        related_abjad_leaves_per_simple_event: typing.Tuple[
            typing.Tuple[typing.Tuple[int, ...], ...], ...
        ],
        attachments_per_type_per_event: typing.Tuple[
            typing.Tuple[typing.Optional[abjad_attachments.AbjadAttachment], ...], ...
        ],
    ) -> None:
        for attachments in attachments_per_type_per_event:
            previous_attachment = None
            for related_abjad_leaves_indices, attachment in zip(
                related_abjad_leaves_per_simple_event, attachments
            ):
                if attachment and attachment.is_active:
                    abjad_leaves = tuple(
                        tools.get_nested_item_from_indices(
                            indices,
                            quanitisized_abjad_leaves,
                        )
                        for indices in related_abjad_leaves_indices
                    )
                    processed_abjad_leaves = attachment.process_leaves(
                        abjad_leaves, previous_attachment
                    )
                    for processed_abjad_leaf, indices in zip(
                        processed_abjad_leaves, related_abjad_leaves_indices
                    ):
                        tools.set_nested_item_from_indices(
                            indices, quanitisized_abjad_leaves, processed_abjad_leaf
                        )

                    previous_attachment = attachment

    def _extract_data_from_simple_event(
        self, simple_event: events.basic.SimpleEvent
    ) -> typing.Tuple[
        typing.List[parameters.abc.Pitch],
        parameters.abc.Volume,
        parameters.playing_indicators.PlayingIndicatorCollection,
        parameters.notation_indicators.NotationIndicatorCollection,
    ]:
        try:
            pitches = self._simple_event_to_pitches(simple_event)
        except AttributeError:
            pitches = []

        # TODO(Add option: no dynamic indicator if there aren't any pitches)
        try:
            if pitches:
                volume = self._simple_event_to_volume(simple_event)
            else:
                volume = parameters.volumes.DirectVolume(0)
        except AttributeError:
            volume = parameters.volumes.DirectVolume(0)
            pitches = []

        try:
            playing_indicators = self._simple_event_to_playing_indicators(simple_event)
        except AttributeError:
            playing_indicators = (
                events.music_constants.DEFAULT_PLAYING_INDICATORS_COLLECTION_CLASS()
            )

        try:
            notation_indicators = self._simple_event_to_notation_indicators(
                simple_event
            )
        except AttributeError:
            notation_indicators = (
                events.music_constants.DEFAULT_NOTATION_INDICATORS_COLLECTION_CLASS()
            )

        return pitches, volume, playing_indicators, notation_indicators

    def _apply_pitches_on_quantized_abjad_leaf(
        self,
        quanitisized_abjad_leaves: abjad.Voice,
        abjad_pitches: typing.List[abjad.Pitch],
        related_abjad_leaves_indices: typing.Tuple[typing.Tuple[int, ...], ...],
    ):
        if len(abjad_pitches) == 1:
            leaf_class = abjad.Note
        else:
            leaf_class = abjad.Chord

        for related_abjad_leaf_indices in related_abjad_leaves_indices:
            abjad_leaf = tools.get_nested_item_from_indices(
                related_abjad_leaf_indices, quanitisized_abjad_leaves
            )
            if leaf_class == abjad.Note:
                # skip don't have note heads
                if hasattr(abjad_leaf, "note_head"):
                    abjad_leaf.note_head._written_pitch = abjad_pitches[0]

            else:
                new_abjad_leaf = leaf_class(
                    [abjad.NamedPitch() for _ in abjad_pitches],
                    abjad_leaf.written_duration,
                )
                for indicator in abjad.get.indicators(abjad_leaf):
                    if type(indicator) != dict:
                        abjad.attach(indicator, new_abjad_leaf)

                for abjad_pitch, note_head in zip(
                    abjad_pitches, new_abjad_leaf.note_heads
                ):
                    note_head._written_pitch = abjad_pitch

                tools.set_nested_item_from_indices(
                    related_abjad_leaf_indices,
                    quanitisized_abjad_leaves,
                    new_abjad_leaf,
                )

    def _apply_pitches_on_quantized_abjad_leaves(
        self,
        quanitisized_abjad_leaves: abjad.Voice,
        related_abjad_leaves_per_simple_event: typing.Tuple[
            typing.Tuple[typing.Tuple[int, ...], ...], ...
        ],
        extracted_data_per_simple_event: typing.Tuple[
            typing.Tuple[
                typing.List[parameters.abc.Pitch],
                parameters.abc.Volume,
                parameters.playing_indicators.PlayingIndicatorCollection,
                parameters.notation_indicators.NotationIndicatorCollection,
            ],
            ...,
        ],
        is_simple_event_rest_per_simple_event: typing.Tuple[bool, ...],
    ):
        for is_simple_event_rest, extracted_data, related_abjad_leaves_indices in zip(
            is_simple_event_rest_per_simple_event,
            extracted_data_per_simple_event,
            related_abjad_leaves_per_simple_event,
        ):
            if not is_simple_event_rest:
                pitches = extracted_data[0]
                abjad_pitches = [
                    self._mutwo_pitch_to_abjad_pitch_converter.convert(pitch)
                    for pitch in pitches
                ]
                self._apply_pitches_on_quantized_abjad_leaf(
                    quanitisized_abjad_leaves,
                    abjad_pitches,
                    related_abjad_leaves_indices,
                )

    def _quantize_sequential_event(
        self,
        sequential_event_to_convert: events.basic.SequentialEvent[
            events.basic.SimpleEvent
        ],
        is_simple_event_rest_per_simple_event: typing.Tuple[bool, ...],
    ) -> typing.Tuple[
        abjad.Container,
        typing.Tuple[typing.Tuple[typing.Tuple[int, ...], ...], ...],
    ]:
        is_simple_event_rest_per_simple_event_iterator = iter(
            is_simple_event_rest_per_simple_event
        )
        (
            quanitisized_abjad_leaves,
            related_abjad_leaves_per_simple_event,
        ) = self._sequential_event_to_quantized_abjad_container_converter.convert(
            sequential_event_to_convert.set_parameter(  # type: ignore
                "is_rest",
                lambda _: next(is_simple_event_rest_per_simple_event_iterator),
                set_unassigned_parameter=True,
                mutate=False,
            )
        )
        return quanitisized_abjad_leaves, related_abjad_leaves_per_simple_event

    def _fill_abjad_container(
        self,
        abjad_container_to_fill: abjad.Voice,
        sequential_event_to_convert: events.basic.SequentialEvent[
            events.basic.SimpleEvent
        ],
    ):
        # tie rests before processing the event!
        sequential_event_to_convert = sequential_event_to_convert.tie_by(
            lambda event0, event1: self._is_simple_event_rest(event0)
            and self._is_simple_event_rest(event1),
            event_type_to_examine=events.basic.SimpleEvent,
            mutate=False,
        )

        # first, extract data from simple events and find rests
        extracted_data_per_simple_event = tuple(
            self._extract_data_from_simple_event(simple_event)
            for simple_event in sequential_event_to_convert
        )
        is_simple_event_rest_per_simple_event = tuple(
            self._is_simple_event_rest(simple_event)
            for simple_event in sequential_event_to_convert
        )

        # second, quantize the sequential event
        (
            quanitisized_abjad_leaves,
            related_abjad_leaves_per_simple_event,
        ) = self._quantize_sequential_event(
            sequential_event_to_convert, is_simple_event_rest_per_simple_event
        )

        # third, apply pitches on Abjad voice
        self._apply_pitches_on_quantized_abjad_leaves(
            quanitisized_abjad_leaves,
            related_abjad_leaves_per_simple_event,
            extracted_data_per_simple_event,
            is_simple_event_rest_per_simple_event,
        )

        # fourth, apply dynamics, tempos and playing_indicators on abjad voice
        attachments_per_type_per_event = (
            self._get_attachments_for_quantized_abjad_leaves(
                extracted_data_per_simple_event
            )
        )
        self._apply_attachments_on_quantized_abjad_leaves(
            quanitisized_abjad_leaves,
            related_abjad_leaves_per_simple_event,
            attachments_per_type_per_event,
        )
        self._apply_tempos_on_quantized_abjad_leaves(quanitisized_abjad_leaves)

        # fifth, replace rests lasting one bar with full measure rests
        if self._write_multimeasure_rests:
            SequentialEventToAbjadVoiceConverter._replace_rests_with_full_measure_rests(
                quanitisized_abjad_leaves
            )

        # move leaves from 'quanitisized_abjad_leaves' object to target container
        abjad.mutate.swap(quanitisized_abjad_leaves, abjad_container_to_fill)

    # ###################################################################### #
    #               public methods for interaction with the user             #
    # ###################################################################### #

    def convert(
        self,
        sequential_event_to_convert: events.basic.SequentialEvent[
            events.basic.SimpleEvent
        ],
    ) -> abjad.Voice:
        """Convert passed :class:`~mutwo.events.basic.SequentialEvent`.

        :param sequential_event_to_convert: The
            :class:`~mutwo.events.basic.SequentialEvent` which shall
            be converted to the :class:`abjad.Voice` object.
        :type sequential_event_to_convert: mutwo.events.basic.SequentialEvent

        **Example:**

        >>> import abjad
        >>> from mutwo.events import basic, music
        >>> from mutwo.converters.frontends import abjad as mutwo_abjad
        >>> mutwo_melody = basic.SequentialEvent(
        >>>     [
        >>>         music.NoteLike(pitch, duration)
        >>>         for pitch, duration in zip("c a g e".split(" "), (1, 1 / 6, 1 / 6, 1 / 6))
        >>>     ]
        >>> )
        >>> converter = mutwo_abjad.SequentialEventToAbjadVoiceConverter()
        >>> abjad_melody = converter.convert(mutwo_melody)
        >>> abjad.lilypond(abjad_melody)
        \\new Voice
        {
            {
                \\tempo 4=120
                %%% \\time 4/4 %%%
                c'1
                \\mf
            }
            {
                \\times 2/3 {
                    a'4
                    g'4
                    e'4
                }
                r2
            }
        }
        """

        return super().convert(sequential_event_to_convert)


class NestedComplexEventToComplexEventToAbjadContainerConvertersConverter(
    converters_abc.Converter
):
    @abc.abstractmethod
    def convert(
        self, nested_complex_event_to_convert: events.abc.ComplexEvent
    ) -> typing.Tuple[ComplexEventToAbjadContainerConverter, ...]:
        raise NotImplementedError


class CycleBasedNestedComplexEventToComplexEventToAbjadContainerConvertersConverter(
    NestedComplexEventToComplexEventToAbjadContainerConvertersConverter
):
    def __init__(
        self,
        complex_event_to_abjad_container_converters: typing.Sequence[
            ComplexEventToAbjadContainerConverter
        ],
    ):
        self._complex_event_to_abjad_container_converters = (
            complex_event_to_abjad_container_converters
        )

    def convert(
        self, nested_complex_event_to_convert: events.abc.ComplexEvent
    ) -> typing.Tuple[ComplexEventToAbjadContainerConverter, ...]:
        complex_event_to_abjad_container_converters_cycle = itertools.cycle(
            self._complex_event_to_abjad_container_converters
        )
        complex_event_to_abjad_container_converters = []
        for _ in nested_complex_event_to_convert:
            complex_event_to_abjad_container_converters.append(
                next(complex_event_to_abjad_container_converters_cycle)
            )
        return tuple(complex_event_to_abjad_container_converters)


class TagBasedNestedComplexEventToComplexEventToAbjadContainerConvertersConverter(
    NestedComplexEventToComplexEventToAbjadContainerConvertersConverter
):
    def __init__(
        self,
        tag_to_complex_event_to_abjad_container_converter: typing.Dict[
            str, ComplexEventToAbjadContainerConverter
        ],
        complex_event_to_tag: typing.Callable[
            [events.abc.ComplexEvent], str
        ] = lambda complex_event: complex_event.tag,
    ):
        self._tag_to_complex_event_to_abjad_container_converter = (
            tag_to_complex_event_to_abjad_container_converter
        )
        self._complex_event_to_tag = complex_event_to_tag

    def convert(
        self, nested_complex_event_to_convert: events.abc.ComplexEvent
    ) -> typing.Tuple[ComplexEventToAbjadContainerConverter, ...]:
        complex_event_to_abjad_container_converters = []
        for complex_event in nested_complex_event_to_convert:
            tag = self._complex_event_to_tag(complex_event)
            try:
                complex_event_to_abjad_container_converter = (
                    self._tag_to_complex_event_to_abjad_container_converter[tag]
                )
            except KeyError:
                raise KeyError(
                    f"Found undefined tag '{tag}'."
                    " This object only knows the following tags:"
                    f" '{self._tag_to_complex_event_to_abjad_container_converter.keys()}'"
                )

            complex_event_to_abjad_container_converters.append(
                complex_event_to_abjad_container_converter
            )
        return tuple(complex_event_to_abjad_container_converters)


class NestedComplexEventToAbjadContainerConverter(
    ComplexEventToAbjadContainerConverter
):
    def __init__(
        self,
        nested_complex_event_to_complex_event_to_abjad_container_converters_converter: NestedComplexEventToComplexEventToAbjadContainerConvertersConverter,
        abjad_container_class: typing.Type[abjad.Container],
        lilypond_type_of_abjad_container: str,
        complex_event_to_abjad_container_name: typing.Callable[
            [events.abc.ComplexEvent], str
        ] = lambda complex_event: complex_event.tag,
        pre_process_abjad_container_routines: typing.Sequence[
            abjad_process_container_routines.ProcessAbjadContainerRoutine
        ] = tuple([]),
        post_process_abjad_container_routines: typing.Sequence[
            abjad_process_container_routines.ProcessAbjadContainerRoutine
        ] = tuple([]),
    ):
        super().__init__(
            abjad_container_class,
            lilypond_type_of_abjad_container,
            complex_event_to_abjad_container_name,
            pre_process_abjad_container_routines,
            post_process_abjad_container_routines,
        )
        self._nested_complex_event_to_complex_event_to_abjad_container_converters_converter = nested_complex_event_to_complex_event_to_abjad_container_converters_converter

    def _fill_abjad_container(
        self,
        abjad_container_to_fill: abjad.Container,
        nested_complex_event_to_convert: events.abc.ComplexEvent,
    ):
        complex_event_to_abjad_container_converters = self._nested_complex_event_to_complex_event_to_abjad_container_converters_converter.convert(
            nested_complex_event_to_convert
        )
        for complex_event, complex_event_to_abjad_container_converter in zip(
            nested_complex_event_to_convert, complex_event_to_abjad_container_converters
        ):
            converted_complex_event = (
                complex_event_to_abjad_container_converter.convert(complex_event)
            )
            abjad_container_to_fill.append(converted_complex_event)



"""Module to convert mutwo parameters to abjad equivalents."""

import abc
import typing

try:
    import quicktions as fractions  # type: ignore
except ImportError:
    import fractions  # type: ignore

import abjad  # type: ignore
import expenvelope  # type: ignore

from mutwo.converters import abc as converters_abc
from mutwo.converters.frontends import abjad_attachments
from mutwo.converters.frontends import ekmelily_constants

from mutwo import parameters

from mutwo.utilities import constants
from mutwo.utilities import tools

__all__ = (
    "MutwoPitchToAbjadPitchConverter",
    "MutwoPitchToHEJIAbjadPitchConverter",
    "MutwoVolumeToAbjadAttachmentDynamicConverter",
    "TempoEnvelopeToAbjadAttachmentTempoConverter",
    "ComplexTempoEnvelopeToAbjadAttachmentTempoConverter",
)


class MutwoPitchToAbjadPitchConverter(converters_abc.Converter):
    """Convert Mutwo Pitch objects to Abjad Pitch objects.

    This default class simply checks if the passed Mutwo object belongs to
    :class:`mutwo.parameters.pitches.WesternPitch`. If it does, Mutwo
    will initialise the Abjad Pitch from the :attr:`name` attribute.
    Otherwise Mutwo will simply initialise the Abjad Pitch from the
    objects :attr:`frequency` attribute.

    If users desire to make more complex conversions (for instance
    due to ``scordatura`` or transpositions of instruments), one can simply
    inherit from this class to define more complex cases.
    """

    def convert(self, pitch_to_convert: parameters.abc.Pitch) -> abjad.Pitch:
        if isinstance(pitch_to_convert, parameters.pitches.WesternPitch):
            return abjad.NamedPitch(pitch_to_convert.name)
        else:
            return abjad.NamedPitch.from_hertz(pitch_to_convert.frequency)


class MutwoPitchToHEJIAbjadPitchConverter(MutwoPitchToAbjadPitchConverter):
    """Convert Mutwo :obj:`~mutwo.parameters.pitches.JustIntonationPitch` objects to Abjad Pitch objects.

    :param reference_pitch: The reference pitch (1/1). Should be a diatonic
        pitch name (see
        :const:`~mutwo.parameters.pitches_constants.ASCENDING_DIATONIC_PITCH_NAMES`)
        in English nomenclature. For any other reference pitch than 'c', Lilyponds
        midi rendering for pitches with the diatonic pitch 'c' will be slightly
        out of tune (because the first value of :arg:`global_scale`
        always have to be 0).
    :type reference_pitch: str, optional
    :param prime_to_heji_accidental_name: Mapping of a prime number
        to a string which indicates the respective prime number in the resulting
        accidental name. See
        :const:`~mutwo.converters.frontends.ekmelily_constants.DEFAULT_PRIME_TO_HEJI_ACCIDENTAL_NAME`
        for the default mapping.
    :type prime_to_heji_accidental_name: typing.Dict[int, str], optional
    :param otonality_indicator: String which indicates that the
        respective prime alteration is otonal. See
        :const:`~mutwo.converters.frontends.ekmelily_constants.DEFAULT_OTONALITY_INDICATOR`
        for the default value.
    :type otonality_indicator: str, optional
    :param utonality_indicator: String which indicates that the
        respective prime alteration is utonal. See
        :const:`~mutwo.converters.frontends.ekmelily_constants.DEFAULT_OTONALITY_INDICATOR`
        for the default value.
    :type utonality_indicator: str, optional
    :param exponent_to_exponent_indicator: Function to convert the
        exponent of a prime number to string which indicates the respective
        exponent. See
        :func:`~mutwo.converters.frontends.ekmelily_constants.DEFAULT_EXPONENT_TO_EXPONENT_INDICATOR`
        for the default function.
    :type exponent_to_exponent_indicator: typing.Callable[[int], str], optional
    :param tempered_pitch_indicator: String which indicates that the
        respective accidental is tempered (12 EDO). See
        :const:`~mutwo.converters.frontends.ekmelily_constants.DEFAULT_TEMPERED_PITCH_INDICATOR`
        for the default value.
    :type tempered_pitch_indicator: str, optional

    The resulting Abjad pitches are expected to be used in combination with tuning
    files that are generated by
    :class:`~mutwo.converters.frontends.ekmelily.HEJIEkmelilyTuningFileConverter`
    and with the Lilypond extension
    `Ekmelily <http://www.ekmelic-music.org/en/extra/ekmelily.htm>`_.
    You can find pre-generated tuning files
    `here <https://github.com/levinericzimmermann/ekme-heji.ily>`_.

    **Example:**

    >>> from mutwo.parameters import pitches
    >>> from mutwo.converters.frontends import abjad
    >>> my_ji_pitch = pitches.JustIntonationPitch('5/4')
    >>> converter_on_a = abjad.MutwoPitchToHEJIAbjadPitchConverter(reference_pitch='a')
    >>> converter_on_c = abjad.MutwoPitchToHEJIAbjadPitchConverter(reference_pitch='c')
    >>> converter_on_a.convert(my_ji_pitch)
    NamedPitch("csoaa''")
    >>> converter_on_c.convert(my_ji_pitch)
    NamedPitch("eoaa'")
    """

    class _HEJIAccidental(object):
        """Fake abjad accidental

        Only for internal usage within the :class:`MutwoPitchToHEJIAbjadPitchConverter`.
        """

        def __init__(self, accidental: str):
            self._accidental = accidental

        def __str__(self) -> str:
            return self._accidental

        # necessary attributes, although they
        # won't be used at all
        semitones = 0
        arrow = None

    def __init__(
        self,
        reference_pitch: str = "a",
        prime_to_heji_accidental_name: typing.Optional[typing.Dict[int, str]] = None,
        otonality_indicator: str = None,
        utonality_indicator: str = None,
        exponent_to_exponent_indicator: typing.Callable[[int], str] = None,
        tempered_pitch_indicator: str = None,
    ):
        # set default values
        if prime_to_heji_accidental_name is None:
            prime_to_heji_accidental_name = (
                ekmelily_constants.DEFAULT_PRIME_TO_HEJI_ACCIDENTAL_NAME
            )

        if otonality_indicator is None:
            otonality_indicator = ekmelily_constants.DEFAULT_OTONALITY_INDICATOR

        if utonality_indicator is None:
            utonality_indicator = ekmelily_constants.DEFAULT_UTONALITY_INDICATOR

        if exponent_to_exponent_indicator is None:
            exponent_to_exponent_indicator = (
                ekmelily_constants.DEFAULT_EXPONENT_TO_EXPONENT_INDICATOR
            )

        if tempered_pitch_indicator is None:
            tempered_pitch_indicator = (
                ekmelily_constants.DEFAULT_TEMPERED_PITCH_INDICATOR
            )

        self._reference_pitch = reference_pitch
        self._otonality_indicator = otonality_indicator
        self._utonality_indicator = utonality_indicator
        self._exponent_to_exponent_indicator = exponent_to_exponent_indicator
        self._tempered_pitch_indicator = tempered_pitch_indicator
        self._reference_index = (
            parameters.pitches_constants.ASCENDING_DIATONIC_PITCH_NAMES.index(
                reference_pitch
            )
        )
        self._prime_to_heji_accidental_name = prime_to_heji_accidental_name

    def _find_western_octave_for_just_intonation_pitch(
        self,
        pitch_to_convert: parameters.pitches.JustIntonationPitch,
        closest_pythagorean_pitch_name: str,
    ) -> int:
        octave = pitch_to_convert.octave + 4
        closest_pythagorean_pitch_index = (
            parameters.pitches_constants.ASCENDING_DIATONIC_PITCH_NAMES.index(
                closest_pythagorean_pitch_name[0]
            )
        )
        if closest_pythagorean_pitch_index < self._reference_index:
            octave += 1

        pitch_as_western_pitch = parameters.pitches.WesternPitch(
            closest_pythagorean_pitch_name[0], octave
        )
        reference_pitch_as_western_pitch = parameters.pitches.WesternPitch(
            self._reference_pitch, 4
        )
        expected_difference_in_cents = pitch_to_convert.cents
        while (
            expected_difference_in_cents
            - (
                (
                    pitch_as_western_pitch.midi_pitch_number
                    - reference_pitch_as_western_pitch.midi_pitch_number
                )
                * 100
            )
            > 300
        ):
            pitch_as_western_pitch.octave += 1

        while (
            expected_difference_in_cents
            - (
                (
                    pitch_as_western_pitch.midi_pitch_number
                    - reference_pitch_as_western_pitch.midi_pitch_number
                )
                * 100
            )
            < -300
        ):
            pitch_as_western_pitch.octave -= 1

        """
        # for pitches which are written with the same diatonic pitch as
        # the reference_pitch, but which are slightly deeper
        if (
            closest_pythagorean_pitch_index == self._reference_index
            and pitch_to_convert.normalize(mutate=False).cents > 600  # type: ignore
        ):
            octave += 1
        print(octave, closest_pythagorean_pitch_index, self._reference_index, '\n')

        return octave
        """
        return pitch_as_western_pitch.octave

    def _find_heji_accidental_for_just_intonation_pitch(
        self,
        pitch_to_convert: parameters.pitches.JustIntonationPitch,
        abjad_pitch_class: abjad.NamedPitchClass,
    ):
        # find additional commas
        accidental_parts = [str(abjad_pitch_class.accidental)]
        prime_to_exponent = (
            pitch_to_convert.helmholtz_ellis_just_intonation_notation_commas.prime_to_exponent
        )
        for prime in sorted(prime_to_exponent.keys()):
            exponent = prime_to_exponent[prime]
            if exponent != 0:
                tonality = (
                    self._otonality_indicator
                    if exponent > 0
                    else self._utonality_indicator
                )
                heji_accidental_name = self._prime_to_heji_accidental_name[prime]
                exponent_indicator = self._exponent_to_exponent_indicator(
                    abs(exponent) - 1
                )
                accidental_parts.append(
                    f"{tonality}{heji_accidental_name}{exponent_indicator}"
                )

        accidental = self._HEJIAccidental("".join(accidental_parts))
        return accidental

    def _convert_just_intonation_pitch(
        self,
        pitch_to_convert: parameters.pitches.JustIntonationPitch,
    ) -> abjad.Pitch:
        # find pythagorean pitch
        closest_pythagorean_pitch_name = (
            pitch_to_convert.get_closest_pythagorean_pitch_name(self._reference_pitch)
        )
        abjad_pitch_class = abjad.NamedPitchClass(closest_pythagorean_pitch_name)

        accidental = self._find_heji_accidental_for_just_intonation_pitch(
            pitch_to_convert, abjad_pitch_class
        )
        abjad_pitch_class._accidental = accidental

        octave = self._find_western_octave_for_just_intonation_pitch(
            pitch_to_convert, closest_pythagorean_pitch_name
        )

        abjad_pitch = abjad.NamedPitch(octave=octave)
        abjad_pitch._pitch_class = abjad_pitch_class
        return abjad_pitch

    def convert(self, pitch_to_convert: parameters.abc.Pitch) -> abjad.Pitch:
        if isinstance(pitch_to_convert, parameters.pitches.JustIntonationPitch):
            abjad_pitch = self._convert_just_intonation_pitch(pitch_to_convert)
        else:
            abjad_pitch = MutwoPitchToAbjadPitchConverter().convert(pitch_to_convert)

        return abjad_pitch


class MutwoVolumeToAbjadAttachmentDynamicConverter(converters_abc.Converter):
    """Convert Mutwo Volume objects to :class:`~mutwo.converters.frontends.abjad_attachments.Dynamic`.

    This default class simply checks if the passed Mutwo object belongs to
    :class:`mutwo.parameters.volumes.WesternVolume`. If it does, Mutwo
    will initialise the :class:`Tempo` object from the :attr:`name` attribute.
    Otherwise Mutwo will first initialise a :class:`WesternVolume` object via
    its py:method:`mutwo.parameters.volumes.WesternVolume.from_amplitude` method.

    Hairpins aren't notated with the aid of :class:`mutwo.parameters.abc.Volume`
    objects, but with :class:`mutwo.parameters.playing_indicators.Hairpin`.
    """

    def convert(
        self, volume_to_convert: parameters.abc.Volume
    ) -> typing.Optional[abjad_attachments.Dynamic]:
        if not isinstance(volume_to_convert, parameters.volumes.WesternVolume):
            if volume_to_convert.amplitude > 0:
                volume_to_convert = parameters.volumes.WesternVolume.from_amplitude(
                    volume_to_convert.amplitude
                )
            else:
                return None
        return abjad_attachments.Dynamic(dynamic_indicator=volume_to_convert.name)


class TempoEnvelopeToAbjadAttachmentTempoConverter(converters_abc.Converter):
    """Convert tempo envelope to :class:`~mutwo.converters.frontends.abjad_attachments.Tempo`.

    Abstract base class for tempo envelope conversion. See
    :class:`ComplexTempoEnvelopeToAbjadAttachmentTempoConverter` for a concrete
    class.
    """

    @abc.abstractmethod
    def convert(
        self, tempo_envelope_to_convert: expenvelope.Envelope
    ) -> typing.Tuple[typing.Tuple[constants.Real, abjad_attachments.Tempo], ...]:
        # return tuple filled with subtuples (leaf_index, abjad_attachments.Tempo)
        raise NotImplementedError()


class ComplexTempoEnvelopeToAbjadAttachmentTempoConverter(
    TempoEnvelopeToAbjadAttachmentTempoConverter
):
    """Convert tempo envelope to :class:`~mutwo.converters.frontends.abjad_attachments.Tempo`.

    This object tries to intelligently set correct tempo attachments to an
    :class:`abjad.Voice` object, appropriate to Western notation standards.
    Therefore it will not repeat tempo indications if they are merely repetitions
    of previous tempo indications and it will write 'a tempo' when returning to the
    same tempo after ritardandi or accelerandi.
    """

    # ###################################################################### #
    #                     private static methods                             #
    # ###################################################################### #

    @staticmethod
    def _convert_tempo_points(
        tempo_points: typing.Tuple[
            typing.Union[constants.Real, parameters.tempos.TempoPoint], ...
        ]
    ) -> typing.Tuple[parameters.tempos.TempoPoint, ...]:
        return tuple(
            tempo_point
            if isinstance(tempo_point, parameters.tempos.TempoPoint)
            else parameters.tempos.TempoPoint(float(tempo_point))
            for tempo_point in tempo_points
        )

    @staticmethod
    def _find_dynamic_change_indication(
        tempo_point: parameters.tempos.TempoPoint,
        next_tempo_point: typing.Optional[parameters.tempos.TempoPoint],
    ) -> typing.Optional[str]:
        dynamic_change_indication = None
        if next_tempo_point:
            absolute_tempo_for_current_tempo_point = (
                tempo_point.absolute_tempo_in_beat_per_minute
            )
            absolute_tempo_for_next_tempo_point = (
                next_tempo_point.absolute_tempo_in_beat_per_minute
            )
            if (
                absolute_tempo_for_current_tempo_point
                > absolute_tempo_for_next_tempo_point
            ):
                dynamic_change_indication = "rit."
            elif (
                absolute_tempo_for_current_tempo_point
                < absolute_tempo_for_next_tempo_point
            ):
                dynamic_change_indication = "acc."

        return dynamic_change_indication

    @staticmethod
    def _shall_write_metronome_mark(
        tempo_envelope_to_convert: expenvelope.Envelope,
        nth_tempo_point: int,
        tempo_point: parameters.tempos.TempoPoint,
        tempo_points: typing.Tuple[parameters.tempos.TempoPoint, ...],
    ) -> bool:
        write_metronome_mark = True
        for previous_tempo_point, previous_tempo_point_duration in zip(
            reversed(tempo_points[:nth_tempo_point]),
            reversed(tempo_envelope_to_convert.durations[:nth_tempo_point]),
        ):
            # make sure the previous tempo point could have been written
            # down (longer duration than minimal duration)
            if previous_tempo_point_duration > 0:
                # if the previous writeable MetronomeMark has the same
                # beats per minute than the current event, there is no
                # need to write it down again
                if (
                    previous_tempo_point.absolute_tempo_in_beat_per_minute
                    == tempo_point.absolute_tempo_in_beat_per_minute
                ):
                    write_metronome_mark = False
                    break

                # but if it differs, we should definitely write it down
                else:
                    break

        return write_metronome_mark

    @staticmethod
    def _shall_stop_dynamic_change_indication(
        tempo_attachments: typing.Tuple[
            typing.Tuple[constants.Real, abjad_attachments.Tempo], ...
        ]
    ) -> bool:
        stop_dynamic_change_indicaton = False
        for _, previous_tempo_attachment in reversed(tempo_attachments):
            # make sure the previous tempo point could have been written
            # down (longer duration than minimal duration)
            if previous_tempo_attachment.dynamic_change_indication is not None:
                stop_dynamic_change_indicaton = True
            break

        return stop_dynamic_change_indicaton

    @staticmethod
    def _find_metronome_mark_values(
        write_metronome_mark: bool,
        tempo_point: parameters.tempos.TempoPoint,
        stop_dynamic_change_indicaton: bool,
    ) -> typing.Tuple[
        typing.Optional[typing.Tuple[int, int]],
        typing.Optional[typing.Union[int, typing.Tuple[int, int]]],
        typing.Optional[str],
    ]:
        if write_metronome_mark:
            textual_indication: typing.Optional[str] = tempo_point.textual_indication
            reference = fractions.Fraction(tempo_point.reference) * fractions.Fraction(
                1, 4
            )
            reference_duration: typing.Optional[typing.Tuple[int, int]] = (
                reference.numerator,
                reference.denominator,
            )
            units_per_minute: typing.Optional[
                typing.Union[int, typing.Tuple[int, int]]
            ] = (
                (
                    int(tempo_point.tempo_or_tempo_range_in_beats_per_minute[0]),
                    int(tempo_point.tempo_or_tempo_range_in_beats_per_minute[1]),
                )
                if isinstance(
                    tempo_point.tempo_or_tempo_range_in_beats_per_minute, tuple
                )
                else int(tempo_point.tempo_or_tempo_range_in_beats_per_minute)
            )

        else:
            reference_duration = None
            units_per_minute = None
            # check if you can write 'a tempo'
            if stop_dynamic_change_indicaton:
                textual_indication = "a tempo"
            else:
                textual_indication = None

        return reference_duration, units_per_minute, textual_indication

    @staticmethod
    def _process_tempo_event(
        tempo_envelope_to_convert: expenvelope.Envelope,
        nth_tempo_point: int,
        tempo_point: parameters.tempos.TempoPoint,
        tempo_points: typing.Tuple[parameters.tempos.TempoPoint, ...],
        tempo_attachments: typing.Tuple[
            typing.Tuple[constants.Real, abjad_attachments.Tempo], ...
        ],
    ) -> abjad_attachments.Tempo:
        try:
            next_tempo_point: typing.Optional[
                parameters.tempos.TempoPoint
            ] = tempo_points[nth_tempo_point + 1]
        except IndexError:
            next_tempo_point = None

        # check for dynamic_change_indication
        dynamic_change_indication = ComplexTempoEnvelopeToAbjadAttachmentTempoConverter._find_dynamic_change_indication(
            tempo_point, next_tempo_point
        )
        write_metronome_mark = ComplexTempoEnvelopeToAbjadAttachmentTempoConverter._shall_write_metronome_mark(
            tempo_envelope_to_convert,
            nth_tempo_point,
            tempo_point,
            tempo_points,
        )

        stop_dynamic_change_indicaton = ComplexTempoEnvelopeToAbjadAttachmentTempoConverter._shall_stop_dynamic_change_indication(
            tempo_attachments
        )

        (
            reference_duration,
            units_per_minute,
            textual_indication,
        ) = ComplexTempoEnvelopeToAbjadAttachmentTempoConverter._find_metronome_mark_values(
            write_metronome_mark, tempo_point, stop_dynamic_change_indicaton
        )

        # for writing 'a tempo'
        if textual_indication == "a tempo":
            write_metronome_mark = True

        converted_tempo_point = abjad_attachments.Tempo(
            reference_duration=reference_duration,
            units_per_minute=units_per_minute,
            textual_indication=textual_indication,
            dynamic_change_indication=dynamic_change_indication,
            stop_dynamic_change_indicaton=stop_dynamic_change_indicaton,
            print_metronome_mark=write_metronome_mark,
        )

        return converted_tempo_point

    # ###################################################################### #
    #                           public api                                   #
    # ###################################################################### #

    def convert(
        self, tempo_envelope_to_convert: expenvelope.Envelope
    ) -> typing.Tuple[typing.Tuple[constants.Real, abjad_attachments.Tempo], ...]:
        tempo_points = (
            ComplexTempoEnvelopeToAbjadAttachmentTempoConverter._convert_tempo_points(
                tempo_envelope_to_convert.levels
            )
        )

        tempo_attachments: typing.List[
            typing.Tuple[constants.Real, abjad_attachments.Tempo]
        ] = []
        for nth_tempo_point, absolute_time, duration, tempo_point in zip(
            range(len(tempo_points)),
            tools.accumulate_from_zero(tempo_envelope_to_convert.durations),
            tempo_envelope_to_convert.durations + (1,),
            tempo_points,
        ):

            if duration > 0:
                tempo_attachment = ComplexTempoEnvelopeToAbjadAttachmentTempoConverter._process_tempo_event(
                    tempo_envelope_to_convert,
                    nth_tempo_point,
                    tempo_point,
                    tempo_points,
                    tuple(tempo_attachments),
                )
                tempo_attachments.append((absolute_time, tempo_attachment))

        return tuple(tempo_attachments)



"""Generic algorithms which are designed by mutwo authors."""

import numbers
import typing

import expenvelope  # type: ignore
import numpy as np  # type: ignore

from mutwo import utilities

__all__ = ("DynamicChoice",)


class DynamicChoice(object):
    """Weighted random choices with dynamically changing weights.

    :param values: The items to choose from.
    :type values: typing.Iterable[typing.Any]
    :param curves: The dynamically changing weight for each value.
    :type curves: typing.Iterable[expenvelope.Envelope]
    :param random_seed: The seed which shall be set at class initialisation.
    :type random_seed: int
    """

    def __init__(
        self,
        values: typing.Iterable[typing.Any],
        curves: typing.Iterable[expenvelope.Envelope],
        random_seed: int = 100,
    ):

        assert len(values) == len(curves)

        self._values = values
        self._curves = curves
        self._random = np.random.default_rng(random_seed)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._values})"

    def items(self) -> typing.Tuple[typing.Tuple[typing.Any, expenvelope.Envelope]]:
        return tuple(zip(self._values, self._curves))

    def gamble_at(self, time: numbers.Real) -> typing.Any:
        """Return value at requested time.

        :param time: At which position on the x-Axis shall be gambled.
        :type time: numbers.Real
        """
        weights = [curve.value_at(time) for curve in self._curves]
        return self._random.choice(
            self._values, p=utilities.tools.scale_sequence_to_sum(weights, 1)
        )
